import re

from pympi.Praat import TextGrid


class TextHash(dict):
    def __init__(self, sentence_list, transducer):
        data = {}
        for i, sentence in enumerate(sentence_list):
            if sentence:
                data[f"s{i}"] = {"text": transducer(sentence)}
            else:
                continue
            words = sentence.split()
            for j, word in enumerate(words):
                data[f"s{i}w{j}"] = {"text": transducer(word)}
        super().__init__(data)


def create_transducer(text, labels_dictionary, debug=False):
    # deferred expensive imports
    from g2p import make_g2p
    from g2p.mappings import Mapping
    from g2p.transducer import CompositeTransducer, Transducer

    text = text.lower()
    allowable_chars = labels_dictionary.keys()
    fallback_mapping = {}
    und_transducer = make_g2p("und", "und-ascii")
    text = und_transducer(text).output_string
    for char in text:
        if char not in allowable_chars and char not in fallback_mapping:
            fallback_mapping[char] = ""
    for k in fallback_mapping.keys():
        if debug:
            print(
                f"Found {k} which is not modelled by Wav2Vec2; skipping for alignment"
            )
    punctuation_transducer = Transducer(
        Mapping(
            rules=[{"in": re.escape(k), "out": v} for k, v in fallback_mapping.items()],
            in_lang="und-ascii",
            out_lang="uroman",
            case_sensitive=False,
        )
    )
    und_transducer.__setattr__("norm_form", "NFC")
    return CompositeTransducer([und_transducer, punctuation_transducer])


def read_text(text_path):
    with open(text_path) as f:
        return [x.strip() for x in f]


def create_text_grid_from_segments(segments, seg_name, frame_ratio, sample_rate=16000):
    xmax = (frame_ratio * segments[-1].end) / sample_rate
    tg = TextGrid(xmax=xmax)
    value_tier = tg.add_tier(seg_name)
    score_tier = tg.add_tier(f"{seg_name}-score")
    for segment in segments:
        start = (frame_ratio * segment.start) / sample_rate
        end = (frame_ratio * segment.end) / sample_rate
        value_tier.add_interval(start, end, segment.label)
        score_tier.add_interval(start, end, "{:.2f}".format(segment.score))
    return tg



from typing import Optional

import torch
from matplotlib import pyplot as plt

from ..type_definitions_heavy import Stats

BASENAME_MAX_LENGTH = 20


def mask_from_lens(lens, max_len: Optional[int] = None):
    if max_len is None:
        max_len = lens.max()
    ids = torch.arange(0, max_len, device=lens.device, dtype=lens.dtype)
    return torch.lt(ids, lens.unsqueeze(1))


def get_mask_from_lengths(lengths, max_len=None):
    batch_size = lengths.shape[0]
    if max_len is None:
        max_len = max(lengths).item()

    ids = (
        torch.arange(0, max_len).unsqueeze(0).expand(batch_size, -1).to(lengths.device)
    )
    return ids >= lengths.unsqueeze(1).expand(-1, max_len)


def make_positions(tensor, padding_idx):
    """Replace non-padding symbols with their position numbers.
    Position numbers begin at padding_idx+1. Padding symbols are ignored.
    """
    # The series of casts and type-conversions here are carefully
    # balanced to both work with ONNX export and XLA. In particular XLA
    # prefers ints, cumsum defaults to output longs, and ONNX doesn"t know
    # how to handle the dtype kwarg in cumsum.
    mask = tensor.ne(padding_idx).int()
    return (torch.cumsum(mask, dim=1).type_as(mask) * mask).long() + padding_idx


def plot_attn_maps(attn_softs, attn_hards, mel_lens, text_lens, n=4):
    bs = len(attn_softs)
    n = min(n, bs)
    s = bs // n
    attn_softs = attn_softs[::s].cpu().numpy()
    attn_hards = attn_hards[::s].cpu().numpy()
    figs = []
    for attn_soft, attn_hard, mel_len, text_len in zip(
        attn_softs, attn_hards, mel_lens, text_lens
    ):
        attn_soft = attn_soft[:, :mel_len, :text_len].squeeze(0).transpose()
        attn_hard = attn_hard[:, :mel_len, :text_len].squeeze(0).transpose()
        fig, axs = plt.subplots(2, 1)
        axs[0].imshow(attn_soft, aspect="auto", origin="lower")
        axs[1].imshow(attn_hard, aspect="auto", origin="lower")
        fig.canvas.draw()
        figs.append(fig)
    return figs


def plot_mel(data, stats: Stats, titles):
    data_len = len(data)
    fig, axes = plt.subplots(data_len, 1, squeeze=False)
    fig.tight_layout(pad=2.0)
    if titles is None:
        titles = [None for i in range(data_len)]

    def add_axis(fig, old_ax):
        ax = fig.add_axes(old_ax.get_position(), anchor="W")
        ax.set_facecolor("None")
        return ax

    for i in range(data_len):
        mel = data[i]["mel"]
        energy = data[i]["energy"] * stats.energy.std + stats.energy.mean
        pitch = data[i]["pitch"] * stats.pitch.std + stats.pitch.mean
        axes[i][0].imshow(mel, origin="lower")
        axes[i][0].set_aspect(2.5, adjustable="box")
        axes[i][0].set_ylim(0, mel.shape[0])
        axes[i][0].set_title(titles[i], fontsize="medium")
        axes[i][0].tick_params(labelsize="x-small", left=False, labelleft=False)
        axes[i][0].set_anchor("W")

        ax1 = add_axis(fig, axes[i][0])
        ax1.plot(pitch, color="tomato")
        ax1.set_xlim(0, mel.shape[1])
        ax1.set_ylim(0, stats.pitch.max)
        ax1.set_ylabel("F0", color="tomato")
        ax1.tick_params(
            labelsize="x-small", colors="tomato", bottom=False, labelbottom=False
        )

        ax2 = add_axis(fig, axes[i][0])
        ax2.plot(energy, color="darkviolet")
        ax2.set_xlim(0, mel.shape[1])
        ax2.set_ylim(stats.energy.min, stats.energy.max)
        ax2.set_ylabel("Energy", color="darkviolet")
        ax2.yaxis.set_label_position("right")
        ax2.tick_params(
            labelsize="x-small",
            colors="darkviolet",
            bottom=False,
            labelbottom=False,
            left=False,
            labelleft=False,
            right=True,
            labelright=True,
        )

    return fig



"""This is the location of basic cli commands that can be copied to new model cli's
    We use the merge-args to merge function signatures between the base functions described here
    and the model-specific ones defined in everyvoice/model/*/*/*/cli
    We want to do it this way to preserve the functionality from typer's command() decorator
    inferring information from the function signature while still keeping code DRY.
"""

import json
import os
import tempfile
import textwrap
from enum import Enum
from pathlib import Path
from pprint import pformat
from typing import List, Optional, Union

import yaml
from deepdiff import DeepDiff
from loguru import logger
from pydantic import ValidationError
from tqdm import tqdm

from everyvoice.config.type_definitions import TargetTrainingTextRepresentationLevel
from everyvoice.exceptions import InvalidConfiguration
from everyvoice.model.aligner.config import DFAlignerConfig
from everyvoice.model.aligner.DeepForcedAligner.dfaligner.dataset import (
    AlignerDataModule,
)
from everyvoice.model.aligner.DeepForcedAligner.dfaligner.model import Aligner
from everyvoice.model.e2e.config import EveryVoiceConfig
from everyvoice.model.e2e.dataset import E2EDataModule
from everyvoice.model.e2e.model import EveryVoice
from everyvoice.model.feature_prediction.config import FastSpeech2Config
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.dataset import (
    FastSpeech2DataModule,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.model import (
    FastSpeech2,
)
from everyvoice.model.vocoder.config import HiFiGANConfig
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.dataset import (
    HiFiGANDataModule,
)
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.model import HiFiGAN


def load_config_base_command(
    model_config: Union[
        type[DFAlignerConfig],
        type[EveryVoiceConfig],
        type[FastSpeech2Config],
        type[HiFiGANConfig],
    ],
    # Must include the above in model-specific command
    config_args: List[str],
    config_file: Path,
):
    from everyvoice.utils import update_config_from_cli_args

    try:
        config = model_config.load_config_from_path(config_file)
    except ValidationError as error:
        # NOTE: To trigger this error handling code from the command line:
        #   `everyvoice preprocess  config/everyvoice-aligner.yaml`
        import sys

        for config_type in (
            DFAlignerConfig,
            EveryVoiceConfig,
            FastSpeech2Config,
            HiFiGANConfig,
        ):
            try:
                config = config_type.load_config_from_path(  # type: ignore[attr-defined]
                    config_file
                )
                logger.error(
                    f"We are expecting a {model_config.__name__} but it looks like you provided a {config_type.__name__}"
                )
                sys.exit(1)
            except ValidationError:
                pass

        logger.error(f"there was a problem with your config file:\n{error}")
        sys.exit(1)

    config = update_config_from_cli_args(config_args, config)
    return config


def preprocess_base_command(
    model_config: Union[
        type[DFAlignerConfig],
        type[EveryVoiceConfig],
        type[FastSpeech2Config],
        type[HiFiGANConfig],
    ],
    steps: List[str],
    # Must include the above in model-specific command
    config_args: List[str],
    config_file: Path,
    cpus: Optional[int],
    overwrite: bool,
    debug: bool,
):
    from everyvoice.preprocessor import Preprocessor

    config = load_config_base_command(model_config, config_args, config_file)
    preprocessor = Preprocessor(config)
    if (
        isinstance(config, FastSpeech2Config)
        and config.model.target_text_representation_level
        == TargetTrainingTextRepresentationLevel.phonological_features
    ):
        steps.append("pfs")
    preprocessor.preprocess(
        cpus=cpus,
        overwrite=overwrite,
        to_process=steps,
        debug=debug,
    )
    return preprocessor, config, steps


def save_configuration_to_log_dir(
    config: Union[DFAlignerConfig, EveryVoiceConfig, FastSpeech2Config, HiFiGANConfig]
):
    """
    Adds a logging file to the module's logger.
    Records to hparams.yaml the function's configuration.
    """
    log_dir = config.training.logger.save_dir / config.training.logger.name
    log_dir.mkdir(exist_ok=True, parents=True)
    logger.add(log_dir / "log")

    hyperparameters_log = log_dir / "hparams.yaml"
    hyperparameters_log.parent.mkdir(parents=True, exist_ok=True)
    logger.info(
        f"Configuration\n{config.model_dump_json(indent=3)}"
    )  # Once to be logged
    with hyperparameters_log.open(mode="w", encoding="UTF-8") as cout:
        output = json.loads(config.model_dump_json())
        yaml.dump(output, stream=cout)


def train_base_command(
    model_config: Union[
        type[DFAlignerConfig],
        type[EveryVoiceConfig],
        type[FastSpeech2Config],
        type[HiFiGANConfig],
    ],
    data_module: Union[
        type[AlignerDataModule],
        type[E2EDataModule],
        type[FastSpeech2DataModule],
        type[HiFiGANDataModule],
    ],
    model: Union[type[Aligner], type[EveryVoice], type[FastSpeech2], type[HiFiGAN]],
    monitor: str,
    # Must include the above in model-specific command
    config_args: List[str],
    config_file: Path,
    accelerator: str,
    devices: str,
    nodes: int,
    strategy: str,
    gradient_clip_val: float | None,
    model_kwargs={},
):
    from everyvoice.base_cli.callback import ResetValidationDataloaderCallback

    config = load_config_base_command(model_config, config_args, config_file)

    save_configuration_to_log_dir(config)

    logger.info("Loading modules for training...")
    pbar = tqdm(range(4))
    pbar.set_description("Loading pytorch and friends")
    from pytorch_lightning import Trainer

    pbar.update()
    pbar.refresh()
    from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint

    pbar.update()
    pbar.refresh()
    from pytorch_lightning.loggers import TensorBoardLogger

    pbar.update()
    pbar.refresh()
    pbar.set_description("Loading EveryVoice modules")

    pbar.update()
    pbar.refresh()
    tensorboard_logger = TensorBoardLogger(
        **{
            **(config.training.logger.model_dump(exclude={"sub_dir_callable": True})),
            **{"sub_dir": config.training.logger.sub_dir},
        }
    )

    lr_monitor = LearningRateMonitor(logging_interval="step")
    logger.info("Starting training.")
    # This callback will always save the last checkpoint
    # regardless of its performance.
    last_ckpt_callback = ModelCheckpoint(
        save_top_k=1,
        save_last=True,
        every_n_train_steps=config.training.ckpt_steps,
        every_n_epochs=config.training.ckpt_epochs,
        enable_version_counter=True,
    )
    # This callback will only save the top-k checkpoints
    # based on minimization of the monitored loss
    monitored_ckpt_callback = ModelCheckpoint(
        monitor=monitor,
        mode="min",
        save_top_k=config.training.save_top_k_ckpts,
        every_n_train_steps=config.training.ckpt_steps,
        every_n_epochs=config.training.ckpt_epochs,
        enable_version_counter=False,
    )
    trainer = Trainer(
        logger=tensorboard_logger,
        accelerator=accelerator,
        devices=devices,
        max_epochs=config.training.max_epochs,
        max_steps=config.training.max_steps,
        check_val_every_n_epoch=config.training.check_val_every_n_epoch,
        val_check_interval=config.training.val_check_interval,
        callbacks=[
            monitored_ckpt_callback,
            last_ckpt_callback,
            lr_monitor,
            ResetValidationDataloaderCallback(),
        ],
        strategy=strategy,
        num_nodes=nodes,
        detect_anomaly=False,  # used for debugging, but triples training time
        gradient_clip_val=gradient_clip_val,
    )
    data = data_module(config)
    last_ckpt = (
        config.training.finetune_checkpoint
        if config.training.finetune_checkpoint is not None
        and os.path.exists(config.training.finetune_checkpoint)
        else None
    )
    # Train from Scratch
    if last_ckpt is None:
        model_obj = model(config, **model_kwargs)
        logger.info(f"Model's architecture\n{model_obj}")
        tensorboard_logger.log_hyperparams(config.model_dump())
        trainer.fit(model_obj, data)
    else:
        model_obj = model.load_from_checkpoint(last_ckpt)
        logger.info(f"Model's architecture\n{model_obj}")
        # Check if the trainer has changed (but ignore subdir since it is specific to the run)
        if isinstance(model_obj, EveryVoice) or isinstance(config, EveryVoiceConfig):
            optimizer_diff = DeepDiff((), ())
            model_diff = DeepDiff((), ())
        else:
            optimizer_diff = DeepDiff(
                model_obj.config.training.optimizer.model_dump(),
                config.training.optimizer.model_dump(),  # type : ignore[reportAttributeAccessIssue]
            )
            model_diff = DeepDiff(
                model_obj.config.model.model_dump(),
                config.model.model_dump(),
            )
        model_config_diff = []
        optimizer_config_diff = []
        if "values_changed" in model_diff:
            model_config_diff += list(model_diff["values_changed"].items())
        if "types_changed" in model_diff:
            model_config_diff += list(model_diff["types_changed"].items())
        if "values_changed" in optimizer_diff:
            optimizer_config_diff += list(optimizer_diff["values_changed"].items())
        if "types_changed" in optimizer_diff:
            optimizer_config_diff += list(optimizer_diff["types_changed"].items())
        if model_config_diff:
            raise InvalidConfiguration(
                textwrap.dedent(
                    f"""
                    Sorry, you are a trying to fine-tune a model with a different architecture defined in your configuration than was used during pre-training.

                    Please fix your configuration or use a different model.

                    Values Changed: {pformat(model_config_diff)}
                    """
                )
            )
        # If optimizer configuration is different, start training with updated optimizer hyperparameters
        # We need to override the model object's configuration with the current one.
        # This assumes that the model and optimizer configurations haven't changed since they
        # would be caught by the previous checks. TODO: We should also check for certain changes to
        # the text configuration, since certain changes would cause an input space mismatch.
        # FIXME: Cannot assign member "config" for type "HiFiGAN"
        model_obj.config = config
        if hasattr(model_obj, "update_config_settings"):
            model_obj.update_config_settings()
        tensorboard_logger.log_hyperparams(config.model_dump())
        if optimizer_config_diff:
            # Finetune from Checkpoint
            logger.warning(
                textwrap.dedent(
                    f"""
                    Some of your optimizer hyperparameters have changed from your checkpoint at '{last_ckpt}',
                    so we will override your checkpoint hyperparameters and restart the optimizer.

                    Your training logs will start from epoch 0/step 0, but will still use the weights from your checkpoint.

                    Values Changed: {pformat(optimizer_config_diff)}
                    """
                )
            )
            # This will only use the weights in the model_obj, the optimizer and current epoch etc will be restarted using
            # the configuration in model_obj.config, see https://github.com/Lightning-AI/pytorch-lightning/issues/5339
            trainer.fit(model_obj, data)
        else:
            import torch

            logger.info(f"Resuming from checkpoint '{last_ckpt}'")
            # We need to create a temporary checkpoint with torch.save because on_save_checkpoint
            # removes all paths for checkpoint portability. However, some paths, like "vocoder_path"
            # should be still accessible when training is resumed.
            new_config_with_paths = model_obj.config.model_dump(mode="json")
            old_ckpt = torch.load(last_ckpt, map_location=torch.device("cpu"))
            old_ckpt["hyper_parameters"]["config"] = new_config_with_paths
            # TODO: check if we need to do the same thing with stats and any thing else registered on the model
            with tempfile.NamedTemporaryFile() as tmp:
                torch.save(old_ckpt, tmp.name)
                # This will resume the weights, optimizer, and steps from last_ckpt, see https://github.com/Lightning-AI/pytorch-lightning/issues/5339
                trainer.fit(model_obj, data, ckpt_path=tmp.name)


def inference_base_command(_name: Enum):
    pass



from pathlib import Path
from tempfile import TemporaryDirectory
from unittest import TestCase

import torch
from everyvoice.config.shared_types import ContactInformation
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.config import HiFiGANConfig
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.utils import HiFiGAN
from pympi import TextGrid
from pytorch_lightning import Trainer

from ..config import FastSpeech2Config, FastSpeech2TrainingConfig
from ..prediction_writing_callback import (
    PredictionWritingSpecCallback,
    PredictionWritingTextGridCallback,
    PredictionWritingWavCallback,
)
from ..utils import BASENAME_MAX_LENGTH, truncate_basename


class TestTruncateBasename(TestCase):
    """
    Testing truncate_basename().
    """

    def test_short_name(self):
        """
        Short utterances should produce file names that are not truncated.
        """
        output = truncate_basename("Short utterance")
        self.assertEqual(output, "Short-utterance")

    def test_long_name(self):
        """
        Uttrerances longer than BASENAME_MAX_LENGTH should get truncated and
        should have a sha1 in case two utterances have the same prefix.
        """
        output = truncate_basename("A utterance that is too long")
        self.assertEqual(len(output), BASENAME_MAX_LENGTH + 1 + 8)
        self.assertEqual(output, "A-utterance-that-is--d607fba8")

    def test_limit(self):
        """
        Utterances exactly BASENAME_MAX_LENGTH long should not be truncated.
        """
        input = "A" * BASENAME_MAX_LENGTH
        output = truncate_basename(input)
        self.assertEqual(len(output), BASENAME_MAX_LENGTH)
        self.assertEqual(output, input)

    def test_limit_plus_one(self):
        """
        Testing the edge case where the utterance as one too many characters.
        """
        input = "A" * (BASENAME_MAX_LENGTH + 1)
        output = truncate_basename(input)
        self.assertEqual(len(output), BASENAME_MAX_LENGTH + 1 + 8)

    def test_same_prefix_different_names(self):
        """
        Two long utterances with the same prefix should yield different file names.
        """
        prefix = "A" * BASENAME_MAX_LENGTH
        input1 = prefix + "1"
        input2 = prefix + "2"
        output1 = truncate_basename(input1)
        output2 = truncate_basename(input2)
        self.assertNotEqual(output1, output2)
        self.assertRegex(output1, prefix + r"-[0-9A-Fa-f]{8}")
        self.assertRegex(output2, prefix + r"-[0-9A-Fa-f]{8}")


class WritingTestBase(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.contact = ContactInformation(
            contact_name="Test Runner", contact_email="info@everyvoice.ca"
        )
        cls.output_key = "output"
        cls.outputs = {
            cls.output_key: torch.ones([2, 500, 80], device="cpu"),
            "duration_prediction": torch.ones([2, 7], device="cpu"),
            "tgt_lens": [
                90,
                490,
            ],
        }
        cls.batch = {
            "basename": [
                "short",
                "This utterance is way too long",
            ],
            "raw_text": ["test", "W̱SÁNEĆ"],
            "text": [
                torch.IntTensor([2, 3, 4, 5, 6, 7, 8], device="cpu"),
                torch.IntTensor([2, 3, 4, 5, 6, 7, 8], device="cpu"),
            ],
            "speaker": [
                "spk1",
                "spk2",
            ],
            "language": [
                "lngA",
                "lngB",
            ],
        }


class TestWritingSpec(WritingTestBase):
    """
    Testing the callback that writes pt files.
    """

    def test_filenames_not_truncated(self):
        """
        We limit the file name's length to at most BASENAME_MAX_LENGTH in the CLI,
        but the callback does not truncate the basenames passed to it
        """
        with TemporaryDirectory() as tmp_dir:
            tmp_dir = Path(tmp_dir)
            writer = PredictionWritingSpecCallback(
                config=FastSpeech2Config(contact=self.contact),
                global_step=77,
                output_dir=tmp_dir,
                output_key=self.output_key,
            )
            writer.on_predict_batch_end(
                _trainer=None,
                _pl_module=None,
                outputs=self.outputs,
                batch=self.batch,
                _batch_idx=0,
                _dataloader_idx=0,
            )
            output_dir = writer.save_dir
            # print(output_dir, *output_dir.glob("**"))  # For debugging
            self.assertTrue(output_dir.exists())
            self.assertTrue(
                (
                    output_dir / "short--spk1--lngA--spec-pred-22050-mel-librosa.pt"
                ).exists()
            )
            self.assertTrue(
                (
                    output_dir
                    / "This utterance is way too long--spk2--lngB--spec-pred-22050-mel-librosa.pt"
                ).exists()
            )


class TestWritingTextGrid(WritingTestBase):
    """
    Testing the callback that writes TextGrid files.
    """

    def test_filenames_not_truncated(self):
        """
        We limit the file name's length to at most BASENAME_MAX_LENGTH in the CLI,
        but the callback does not truncate the basenames passed to it
        """
        with TemporaryDirectory() as tmp_dir:
            tmp_dir = Path(tmp_dir)
            writer = PredictionWritingTextGridCallback(
                config=FastSpeech2Config(contact=self.contact),
                global_step=77,
                output_dir=tmp_dir,
                output_key=self.output_key,
            )
            writer.on_predict_batch_end(
                _trainer=None,
                _pl_module=None,
                outputs=self.outputs,
                batch=self.batch,
                _batch_idx=0,
                _dataloader_idx=0,
            )
            output_dir = writer.save_dir
            # print(output_dir, *output_dir.glob("**"))  # For debugging
            self.assertTrue(output_dir.exists())
            self.assertTrue(
                (output_dir / "short--spk1--lngA--22050-mel-librosa.TextGrid").exists()
            )
            self.assertTrue(
                (
                    output_dir
                    / "This utterance is way too long--spk2--lngB--22050-mel-librosa.TextGrid"
                ).exists()
            )
            tg = TextGrid(
                file_path=(
                    output_dir
                    / "This utterance is way too long--spk2--lngB--22050-mel-librosa.TextGrid"
                )
            )
            tiers = list(tg.get_tiers())
            self.assertEqual(tiers[0].name, "phones")
            self.assertEqual(tiers[1].name, "phone annotations")
            self.assertEqual(tiers[2].name, "words")
            self.assertEqual(tiers[3].name, "word annotations")
            self.assertEqual(tiers[2].intervals[0][2], "W̱SÁNEĆ")


class TestWritingWav(WritingTestBase):
    """
    Testing the callback that writes wav files.
    Note that this test may be expansive.
    """

    def test_filenames_not_truncated(self):
        """
        We limit the file name's length to at most BASENAME_MAX_LENGTH in the CLI,
        but the callback does not truncate the basenames passed to it
        """
        with TemporaryDirectory() as tmp_dir:
            tmp_dir = Path(tmp_dir)
            contact_info = ContactInformation(
                contact_name="Test Runner", contact_email="info@everyvoice.ca"
            )
            vocoder = HiFiGAN(HiFiGANConfig(contact=contact_info))
            trainer = Trainer(default_root_dir=str(tmp_dir), barebones=True)
            trainer.strategy.connect(vocoder)
            vocoder_path = Path(tmp_dir) / "vocoder"
            trainer.save_checkpoint(vocoder_path)

            writer = PredictionWritingWavCallback(
                config=FastSpeech2Config(
                    contact=self.contact,
                    training=FastSpeech2TrainingConfig(vocoder_path=vocoder_path),
                ),
                device=torch.device("cpu"),
                global_step=77,
                output_dir=tmp_dir,
                output_key=self.output_key,
                vocoder_model=vocoder,
                vocoder_config=vocoder.config,
                vocoder_global_step=10,
            )
            writer.on_predict_batch_end(
                _trainer=None,
                _pl_module=None,
                outputs=self.outputs,
                batch=self.batch,
                _batch_idx=0,
                _dataloader_idx=0,
            )
            output_dir = writer.save_dir
            # print(output_dir, *output_dir.glob("**"))  # For debugging
            self.assertTrue(output_dir.exists())
            self.assertTrue(
                (
                    output_dir / "short--spk1--lngA--ckpt=77--v_ckpt=10--pred.wav"
                ).exists()
            )
            self.assertTrue(
                (
                    output_dir
                    / "This utterance is way too long--spk2--lngB--ckpt=77--v_ckpt=10--pred.wav"
                ).exists()
            )



import json
import os
import sys
from glob import glob
from pathlib import Path

import typer
from everyvoice.base_cli.interfaces import complete_path
from tqdm import tqdm


def audit(
    config_file: Path = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to your model configuration file.",
        autocompletion=complete_path,
    ),
    should_check_stats: bool = True,
    dimensions: bool = True,
):
    import torch

    from ..config import FastSpeech2Config
    from ..type_definitions_heavy import Stats, StatsInfo

    def check_stats(data, path, stats: StatsInfo):
        import torch

        data_min = torch.min(data)
        data_max = torch.max(data)
        assert (
            data_min >= stats.norm_min
        ), f"Data in {path} had min of {data_min} despite stats min being {stats.norm_min}"
        assert (
            data_max <= stats.norm_max
        ), f"Data in {path} had max of {data_min} despite stats max being {stats.norm_max}"

    original_config: FastSpeech2Config = FastSpeech2Config.load_config_from_path(
        config_file
    )
    if should_check_stats:
        with open(original_config.preprocessing.save_dir / "stats.json") as f:
            stats: Stats = Stats(**json.load(f))
    files = []
    for dataset in original_config.preprocessing.source_data:
        files += dataset.filelist_loader(dataset.filelist)
    for x in tqdm(files):
        duration_files = (
            glob(
                os.path.join(
                    (original_config.preprocessing.save_dir / "duration"),
                    f"**/*{x['basename']}*.pt",
                ),
                recursive=True,
            )
            if dimensions
            else []
        )
        energy_files = (
            glob(
                os.path.join(
                    (original_config.preprocessing.save_dir / "energy"),
                    f"**/*{x['basename']}*.pt",
                ),
                recursive=True,
            )
            if dimensions or should_check_stats
            else []
        )
        pitch_files = (
            glob(
                os.path.join(
                    (original_config.preprocessing.save_dir / "pitch"),
                    f"**/*{x['basename']}*.pt",
                ),
                recursive=True,
            )
            if dimensions or should_check_stats
            else []
        )
        text_files = (
            glob(
                os.path.join(
                    (original_config.preprocessing.save_dir / "text"),
                    f"**/*{x['basename']}*.pt",
                ),
                recursive=True,
            )
            if dimensions or should_check_stats
            else []
        )
        if dimensions:
            for dur_path, e_path, p_path, t_path in zip(
                duration_files, energy_files, pitch_files, text_files
            ):
                duration = torch.load(dur_path)
                text = torch.load(t_path)
                assert duration.size(0) == text.size(0)
                e_asserted_duration = (
                    duration.size(0)
                    if original_config.model.variance_predictors.energy.level == "phone"
                    else torch.sum(duration)
                )
                p_asserted_duration = (
                    duration.size(0)
                    if original_config.model.variance_predictors.pitch.level == "phone"
                    else torch.sum(duration)
                )
                e_data = torch.load(e_path)
                assert e_asserted_duration == e_data.size(
                    0
                ), f"Data in {e_path} had duration of {e_data.size(0)} but should have been {e_asserted_duration}"
                if should_check_stats is not None:
                    check_stats(e_data, e_path, stats.energy)
                p_data = torch.load(p_path)
                assert p_asserted_duration == p_data.size(
                    0
                ), f"Data in {p_path} had duration of {p_data.size(0)} but should have been {p_asserted_duration}"
                if stats is not None:
                    check_stats(p_data, p_path, stats.pitch)
        elif should_check_stats:
            for path in energy_files:
                data = torch.load(path)
                check_stats(data, path, stats.energy)
            for path in pitch_files:
                data = torch.load(path)
                check_stats(data, path, stats.pitch)
        else:
            print(
                "Nothing to check. Please re-run with --should_check_stats or --dimensions",
                file=sys.stderr,
            )



# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Source: https://arxiv.org/abs/2108.10447

import numpy as np
from numba import jit, prange


@jit(nopython=True)
def mas(log_attn_map, width=1):
    # assumes mel x text
    opt = np.zeros_like(log_attn_map)
    log_attn_map = log_attn_map.copy()
    log_attn_map[0, 1:] = -np.inf
    log_p = np.zeros_like(log_attn_map)
    log_p[0, :] = log_attn_map[0, :]
    prev_ind = np.zeros_like(log_attn_map, dtype=np.int64)
    for i in range(1, log_attn_map.shape[0]):
        for j in range(log_attn_map.shape[1]):  # for each text dim
            prev_j = np.arange(max(0, j - width), j + 1)
            prev_log = np.array([log_p[i - 1, prev_idx] for prev_idx in prev_j])

            ind = np.argmax(prev_log)
            log_p[i, j] = log_attn_map[i, j] + prev_log[ind]
            prev_ind[i, j] = prev_j[ind]

    # now backtrack
    curr_text_idx = log_attn_map.shape[1] - 1
    for i in range(log_attn_map.shape[0] - 1, -1, -1):
        opt[i, curr_text_idx] = 1
        curr_text_idx = prev_ind[i, curr_text_idx]
    opt[0, curr_text_idx] = 1
    return opt


@jit(nopython=True)
def mas_width1(log_attn_map):
    """mas with hardcoded width=1"""
    # assumes mel x text
    neg_inf = log_attn_map.dtype.type(-np.inf)
    log_p = log_attn_map.copy()
    log_p[0, 1:] = neg_inf
    for i in range(1, log_p.shape[0]):
        prev_log1 = neg_inf
        for j in range(log_p.shape[1]):
            prev_log2 = log_p[i - 1, j]
            log_p[i, j] += max(prev_log1, prev_log2)
            prev_log1 = prev_log2

    # now backtrack
    opt = np.zeros_like(log_p)
    one = opt.dtype.type(1)
    j = log_p.shape[1] - 1
    for i in range(log_p.shape[0] - 1, 0, -1):
        opt[i, j] = one
        if log_p[i - 1, j - 1] >= log_p[i - 1, j]:
            j -= 1
            if j == 0:
                opt[1:i, j] = one
                break
    opt[0, j] = one
    return opt


@jit(nopython=True, parallel=True)
def b_mas(b_log_attn_map, in_lens, out_lens, width=1):
    assert width == 1
    attn_out = np.zeros_like(b_log_attn_map)

    for b in prange(b_log_attn_map.shape[0]):
        out = mas_width1(b_log_attn_map[b, 0, : out_lens[b], : in_lens[b]])
        attn_out[b, 0, : out_lens[b], : in_lens[b]] = out
    return attn_out



"""
This file is for light-weight type definitions with no dependencies.
The file is intended to load in a few milliseconds.
More expensive type definitions belong in shared_types.py.
"""

from enum import Enum


class DatasetTextRepresentation(str, Enum):
    characters = "characters"
    ipa_phones = "phones"
    arpabet = "arpabet"  # always gets mapped to phones


class TargetTrainingTextRepresentationLevel(str, Enum):
    characters = "characters"
    ipa_phones = "phones"
    phonological_features = "phonological_features"



from typing import List
from unicodedata import normalize

import numpy as np
import numpy.typing as npt
from panphon import FeatureTable

from everyvoice.config.text_config import TextConfig

N_PHONOLOGICAL_FEATURES = 39


# TODO: support primary and secondary stress
class PhonologicalFeatureCalculator:
    def __init__(self, text_config: TextConfig, punctuation_hash: dict):
        self.config = text_config
        self.punctuation_hash = punctuation_hash
        self.feature_table = FeatureTable()

    def get_tone_features(self, text: List[str]) -> npt.NDArray[np.float32]:
        # TODO: sort out how to define encoding of tone features
        """Return Wang (1967) style tone features.
            - Contour
            - High
            - Central
            - Mid
            - Rising
            - Falling
            - Convex

        *If your language uses phonemic tone you MUST amend this function to match your language
        Panphon does not use these features.*

        Args:
            text (list(str)): segmented phones
        """
        tone_features = []
        high_tone_chars = [
            normalize("NFC", x)
            for x in [
                "áː",
                "á",
                "ʌ̃́ː",
                "ʌ̃́",
                "éː",
                "é",
                "íː",
                "í",
                "ṹː",
                "ṹ",
                "óː",
                "ó",
            ]
        ]
        low_tone_chars = [
            normalize("NFC", x) for x in ["òː", "ũ̀ː", "ìː", "èː", "ʌ̃̀ː", "àː"]
        ]
        for char in text:
            char = normalize("NFC", char)
            if char in high_tone_chars:
                tone_features.append([-1, 1, -1, -1, -1, -1, -1])
            elif char in low_tone_chars:
                tone_features.append([-1, -1, -1, -1, -1, -1, -1])
            else:
                tone_features.append([0, 0, 0, 0, 0, 0, 0])
        return np.array(tone_features, dtype=np.float32)

    def get_punctuation_features(self, tokens: list[str]) -> npt.NDArray[np.float32]:
        """Get Punctuation features.
           One-hot encodes the allowable types of punctuation and returns zeros elsewhere

        Args:
            tokens (list[str]): a list of IPA and normalized punctuation tokens

        Returns:
            npt.NDArray[np.float32]: a seven-dimensional one-hot encoding of punctuation, white space and silence

        >>> punc_hash = {"exclamations": "<EXCL>", "question_symbols": "<QINT>", "quotemarks": "<QUOTE>", "big_breaks": "<BB>", "small_breaks": "<SB>", "ellipsis": "<EPS>"}
        >>> pf = PhonologicalFeatureCalculator(TextConfig(), punc_hash)
        >>> pf.get_punctuation_features(['h', 'ʌ', 'l', 'o', 'ʊ', '<EXCL>'])
        array([[0., 0., 0., 0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 0., 0., 1., 0.]], dtype=float32)
        """
        punctuation_features = []
        for char in tokens:
            if char == " ":
                punctuation_features.append([1, 0, 0, 0, 0, 0, 0, 0])
            elif char == self.punctuation_hash["question_symbols"]:
                punctuation_features.append([0, 1, 0, 0, 0, 0, 0, 0])
            elif char == self.punctuation_hash["big_breaks"]:
                punctuation_features.append([0, 0, 1, 0, 0, 0, 0, 0])
            elif char == self.punctuation_hash["small_breaks"]:
                punctuation_features.append([0, 0, 0, 1, 0, 0, 0, 0])
            elif char == self.punctuation_hash["quotemarks"]:
                punctuation_features.append([0, 0, 0, 0, 1, 0, 0, 0])
            elif char == self.punctuation_hash["ellipsis"]:
                punctuation_features.append([0, 0, 0, 0, 0, 1, 0, 0])
            elif char == self.punctuation_hash["exclamations"]:
                punctuation_features.append([0, 0, 0, 0, 0, 0, 1, 0])
            elif char in self.config.symbols.silence:
                punctuation_features.append([0, 0, 0, 0, 0, 0, 0, 1])
            else:
                punctuation_features.append([0, 0, 0, 0, 0, 0, 0, 0])
        return np.array(punctuation_features, dtype=np.float32)

    def token_to_segmental_features(self, token: str) -> npt.NDArray[np.float32]:
        """Turn a token to a feature vector with panphon

        Args:
            token (str): a token to convert to segmental features

        Returns:
            npt.NDArray[np.float32]: a list of place and manner of articulation feature values

        >>> punc_hash = {"exclamations": "<EXCL>", "question_symbols": "<QINT>", "quotemarks": "<QUOTE>", "big_breaks": "<BB>", "small_breaks": "<SB>",}
        >>> pf = PhonologicalFeatureCalculator(TextConfig(), punc_hash)
        >>> pf.token_to_segmental_features('\x80') # pad symbol is all zeros
        array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
               0., 0., 0., 0., 0., 0., 0.], dtype=float32)
        >>> pf.token_to_segmental_features('*') # punctuation is all zeros
        array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
               0., 0., 0., 0., 0., 0., 0.], dtype=float32)
        >>> pf.token_to_segmental_features('ʌ')
        array([ 1.,  1., -1.,  1., -1., -1., -1.,  0.,  1., -1., -1.,  0., -1.,
                0., -1., -1., -1.,  1., -1., -1.,  1., -1.,  0.,  0.],
              dtype=float32)
        >>> pf.token_to_segmental_features('a͡ɪ')
        array([ 1.,  1., -1.,  1., -1., -1., -1.,  0.,  1., -1., -1.,  0., -1.,
                0., -1.,  0.,  0., -1., -1., -1.,  0., -1.,  0.,  0.],
              dtype=float32)
        >>> pf.token_to_segmental_features('aɪ')
        array([ 1.,  1., -1.,  1., -1., -1., -1.,  0.,  1., -1., -1.,  0., -1.,
                0., -1.,  0.,  0., -1., -1., -1.,  0., -1.,  0.,  0.],
              dtype=float32)
        """
        vec = self.feature_table.word_to_vector_list(token, numeric=True)
        NUMBER_OF_PANPHON_FEATURES = 24
        # is not IPA
        if not vec:
            return np.zeros(NUMBER_OF_PANPHON_FEATURES, dtype=np.float32)
        # EV and PanPhon tokenization do not match, as with dipthongs
        if len(vec) > 1:
            # TODO: We should warn the user that we are averaging the features due to a mismatch in
            # EV and Panphon tokenization.
            return np.mean(vec, axis=0, dtype=np.float32)
        # EV and PanPhon tokenization matches here
        else:
            return np.array(vec[0], dtype=np.float32)

    def get_features(self, tokens: list[str]) -> npt.NDArray[np.float32]:
        """Get Phonological Feature Vectors by stacking segmental features, tone features, and punctuation features

        Args:
            tokens (list[str]): a list of IPA and normalized punctuation tokens

        Returns:
            npt.NDArray[np.float32]: a thirty-nine-dimensional encoding of segmental (22), tone(7), and punctuation (7) features
        """
        if not tokens:
            return np.array([])
        punctuation_features = self.get_punctuation_features(tokens)
        tone_features = self.get_tone_features(tokens)
        seg_features = np.vstack([self.token_to_segmental_features(t) for t in tokens])
        assert (
            len(punctuation_features) == len(tone_features) == len(seg_features)
        ), "There should be the same number of segments among segmental, tone, and punctuation features"
        return np.concatenate(
            [seg_features, tone_features, punctuation_features],
            axis=1,
            dtype=np.float32,
        )



from importlib import import_module
from pathlib import Path
from typing import Any, Callable, Optional

from loguru import logger
from pydantic import ValidationInfo


def callable_to_string(function: Callable) -> str:
    """Serialize a Callable to a string-formatted Callable"""
    return ".".join([function.__module__, function.__name__])


def string_to_callable(string: str | Callable) -> Callable:
    """De-serialize a string-formatted Callable to a Callable"""
    if callable(string):
        return string
    elif not isinstance(string, str):
        raise ValueError(f"Expected a string or callable, got {type(string)}")
    if "." not in string:
        # Just return a function that returns the string if
        # it's not in the <module>.<function> format
        def curried(*argv, **kwargs):
            return string

        return curried
    module_name, function_name = string.rsplit(".", 1)
    try:
        module = import_module(module_name)
    except ImportError as e:
        raise ImportError(
            f"Cannot import module {module_name} - this must be a valid module"
        ) from e
    try:
        function = getattr(module, function_name)
    except AttributeError as exc:
        raise AttributeError(
            f"Cannot find method {function_name} in module {module}"
        ) from exc
    return function


def relative_to_absolute_path(
    value: Any, info: Optional[ValidationInfo] = None
) -> Path | None:
    """
    Helper function to annotate a type.
    This function processes relative paths and either resolve them to absolute
    paths or resolve them with respect to the configuration file they came
    from.
    """
    if value is None:
        return value

    try:
        # Make sure value is a path because it can be a string when we load a
        # model that is not partial.
        path = Path(value)
        if (
            not path.is_absolute()
            and info
            and info.context
            and (config_path := info.context.get("config_path", None))
        ):
            path = (config_path.parent / path).resolve()
        return path
    except TypeError as e:
        # Pydantic needs ValueErrors to raise its ValidationErrors
        raise ValueError from e


def directory_path_must_exist(
    value: Any, info: Optional[ValidationInfo] = None
) -> Path | None:
    """
    Helper function to annotate a type.
    Creates a directory if it doesn't exist.
    """
    assert isinstance(value, Path)
    if (
        info
        and info.context
        and (writing_config := info.context.get("writing_config", None))
    ):
        # We are writing the original config and must temporarily resolve the path.
        (writing_config.resolve() / value).mkdir(parents=True, exist_ok=True)
    else:
        if not value.exists():
            logger.info(f"Directory at {value} does not exist. Creating...")
            value.mkdir(parents=True, exist_ok=True)

    return value


def path_is_a_directory(
    value: Any, info: Optional[ValidationInfo] = None
) -> Path | None:
    """
    Helper function to annotate a type.
    Verifies ala `PathType("dir")` that `value` is a directory.
    """
    if (
        info
        and info.context
        and (writing_config := info.context.get("writing_config", None))
    ):
        # We are writing the original config and must temporarily resolve the path.
        tmp_path = writing_config.resolve() / value
        if not tmp_path.is_dir():
            raise ValueError(f"{tmp_path} is not a directory")
    else:
        try:
            # Make sure value is a path because it can be a string when we load a model that is not partial.
            path = Path(value)
            if not path.is_dir():
                raise ValueError(f"{path} is not a directory")
        except TypeError as e:
            # Pydantic needs ValueErrors to raise its ValidationErrors
            raise ValueError from e

    return value



import itertools
import math

import pytorch_lightning as pl
import torch
import torch.nn as nn
import torch.nn.functional as F
from everyvoice.model.vocoder.config import VocoderConfig
from everyvoice.utils import plot_spectrogram
from everyvoice.utils.heavy import (
    dynamic_range_compression_torch,
    get_spectral_transform,
)
from torch.nn import AvgPool1d, Conv1d, Conv2d, ConvTranspose1d
from torch.nn.utils import spectral_norm
from torch.nn.utils.parametrizations import weight_norm
from torch.nn.utils.parametrize import remove_parametrizations

from .config import HiFiGANResblock, HiFiGANTrainTypes


def init_weights(m, mean=0.0, std=0.01):
    classname = m.__class__.__name__
    if classname.find("Conv") != -1:
        m.weight.data.normal_(mean, std)
    elif classname.find("Sequential") != -1:
        for layer in m:
            layer.weight.data.normal_(mean, std)


def get_padding(kernel_size, dilation=1):
    return int((kernel_size * dilation - dilation) / 2)


class ResBlock1(torch.nn.Module):
    def __init__(
        self,
        config: VocoderConfig,
        channels,
        kernel_size=3,
        dilation=(1, 3, 5),
    ):
        super(ResBlock1, self).__init__()
        self.config = config
        self.convs1 = nn.ModuleList(
            [
                weight_norm(
                    Conv1d(
                        channels,
                        channels,
                        kernel_size,
                        1,
                        dilation=x,
                        padding=get_padding(kernel_size, x),
                    )
                )
                for x in dilation
            ]
        )
        self.convs1.apply(init_weights)
        self.convs2 = nn.ModuleList(
            [
                weight_norm(
                    Conv1d(
                        channels,
                        channels,
                        kernel_size,
                        1,
                        dilation=1,
                        padding=get_padding(kernel_size, 1),
                    )
                )
                for _ in dilation
            ]
        )
        self.convs2.apply(init_weights)

    def forward(self, x):
        for c1, c2 in zip(self.convs1, self.convs2):
            xt = self.config.model.activation_function(x)
            xt = c1(xt)
            xt = self.config.model.activation_function(xt)
            xt = c2(xt)
            x = xt + x
        return x

    def remove_weight_norm(self):
        for layer in self.convs1:
            if layer.__class__.__name__ == "Sequential":
                for sub_layer in layer:
                    remove_parametrizations(sub_layer, "weight")
            else:
                remove_parametrizations(layer, "weight")
        for layer in self.convs2:
            if layer.__class__.__name__ == "Sequential":
                for sub_layer in layer:
                    remove_parametrizations(sub_layer, "weight")
            else:
                remove_parametrizations(layer, "weight")


class ResBlock2(torch.nn.Module):
    def __init__(
        self,
        config: VocoderConfig,
        channels,
        kernel_size=3,
        dilation=(1, 3),
    ):
        super(ResBlock2, self).__init__()
        self.config = config
        self.convs = nn.ModuleList(
            [
                weight_norm(
                    Conv1d(
                        channels,
                        channels,
                        kernel_size,
                        1,
                        dilation=dilation[0],
                        padding=get_padding(kernel_size, dilation[0]),
                    )
                ),
                weight_norm(
                    Conv1d(
                        channels,
                        channels,
                        kernel_size,
                        1,
                        dilation=dilation[1],
                        padding=get_padding(kernel_size, dilation[1]),
                    )
                ),
            ]
        )
        self.convs.apply(init_weights)

    def forward(self, x):
        for c in self.convs:
            xt = self.config.model.activation_function(x)
            xt = c(xt)
            x = xt + x
        return x

    def remove_weight_norm(self):
        for layer in self.convs:
            if layer.__class__.__name__ == "Sequential":
                for sub_layer in layer:
                    remove_parametrizations(sub_layer, "weight")
            else:
                remove_parametrizations(layer, "weight")


class Generator(torch.nn.Module):
    def __init__(self, config: VocoderConfig):
        super(Generator, self).__init__()
        self.config = config
        self.audio_config = config.preprocessing.audio
        self.sampling_rate_change = (
            self.audio_config.output_sampling_rate
            // self.audio_config.input_sampling_rate
        )
        self.model_vocoder_config = config.model
        self.num_kernels = len(self.model_vocoder_config.resblock_kernel_sizes)
        self.num_upsamples = len(self.model_vocoder_config.upsample_rates)
        self.conv_pre = weight_norm(
            Conv1d(
                self.audio_config.n_mels,
                self.model_vocoder_config.upsample_initial_channel,
                7,
                1,
                padding=3,
            )
        )

        resblock = (
            ResBlock1
            if self.model_vocoder_config.resblock is HiFiGANResblock.one
            else ResBlock2
        )

        self.ups = nn.ModuleList()
        for i, (u, k) in enumerate(
            zip(
                self.model_vocoder_config.upsample_rates,
                self.model_vocoder_config.upsample_kernel_sizes,
            )
        ):
            # TODO: add sensible upsampling layer
            self.ups.append(
                weight_norm(
                    ConvTranspose1d(
                        self.model_vocoder_config.upsample_initial_channel
                        // (2**i),  # in
                        self.model_vocoder_config.upsample_initial_channel
                        // (2 ** (i + 1)),  # out
                        k,  # kernel
                        u,  # stride
                        padding=(k - u) // 2,
                    )
                )
            )

        self.resblocks = nn.ModuleList()
        for i in range(len(self.ups)):
            ch = self.model_vocoder_config.upsample_initial_channel // (2 ** (i + 1))
            for k, d in zip(
                self.model_vocoder_config.resblock_kernel_sizes,
                self.model_vocoder_config.resblock_dilation_sizes,
            ):
                self.resblocks.append(resblock(self.config, ch, k, d))
        if self.config.model.istft_layer:
            self.post_n_fft = (
                self.audio_config.n_fft * self.sampling_rate_change
            ) // math.prod(self.config.model.upsample_rates)
            self.reflection_pad = torch.nn.ReflectionPad1d((1, 0))
            conv_post_out_channels = self.post_n_fft + 2
        else:
            self.post_n_fft = conv_post_out_channels = 1

        self.conv_post = weight_norm(
            Conv1d(ch, conv_post_out_channels, 7, 1, padding=3)
        )
        self.ups.apply(init_weights)
        self.conv_post.apply(init_weights)

    def forward(self, x):
        x = self.conv_pre(x)
        for i in range(self.num_upsamples):
            x = self.config.model.activation_function(x)
            x = self.ups[i](x)
            xs = None
            for j in range(self.num_kernels):
                if xs is None:
                    xs = self.resblocks[i * self.num_kernels + j](x)
                else:
                    xs += self.resblocks[i * self.num_kernels + j](x)
            x = xs / self.num_kernels
        # NOTE: Changed from user provided activation to fixed leaky_relu to mimic our reference code in jik876.
        x = F.leaky_relu(x)
        if self.config.model.istft_layer:
            x = self.reflection_pad(x)
            x = self.conv_post(x)
            spec = torch.exp(x[:, : self.post_n_fft // 2 + 1, :])
            phase = torch.sin(x[:, self.post_n_fft // 2 + 1 :, :])
            return spec, phase
        else:
            x = self.conv_post(x)
            x = torch.tanh(x)
            return x

    def remove_weight_norm(self):
        for layer in self.ups:
            if layer.__class__.__name__ == "Sequential":
                for sub_layer in layer:
                    remove_parametrizations(sub_layer, "weight")
            else:
                remove_parametrizations(layer, "weight")
        for layer in self.resblocks:
            layer.remove_weight_norm()
        if self.conv_pre.__class__.__name__ == "Sequential":
            for sub_layer in self.conv_pre:
                remove_parametrizations(sub_layer, "weight")
        else:
            remove_parametrizations(self.conv_pre, "weight")
        if self.conv_post.__class__.__name__ == "Sequential":
            for sub_layer in self.conv_post:
                remove_parametrizations(sub_layer, "weight")
        else:
            remove_parametrizations(self.conv_post, "weight")


class DiscriminatorP(torch.nn.Module):
    def __init__(
        self, period, config, kernel_size=5, stride=3, use_spectral_norm=False
    ):
        super(DiscriminatorP, self).__init__()
        self.period = period
        self.config = config
        norm_f = weight_norm if use_spectral_norm is False else spectral_norm
        self.convs = nn.ModuleList(
            [
                norm_f(
                    Conv2d(
                        1,
                        32,
                        (kernel_size, 1),
                        (stride, 1),
                        padding=(get_padding(5, 1), 0),
                    )
                ),
                norm_f(
                    Conv2d(
                        32,
                        128,
                        (kernel_size, 1),
                        (stride, 1),
                        padding=(get_padding(5, 1), 0),
                    )
                ),
                norm_f(
                    Conv2d(
                        128,
                        512,
                        (kernel_size, 1),
                        (stride, 1),
                        padding=(get_padding(5, 1), 0),
                    )
                ),
                norm_f(
                    Conv2d(
                        512,
                        1024,
                        (kernel_size, 1),
                        (stride, 1),
                        padding=(get_padding(5, 1), 0),
                    )
                ),
                norm_f(Conv2d(1024, 1024, (kernel_size, 1), 1, padding=(2, 0))),
            ]
        )
        self.conv_post = norm_f(Conv2d(1024, 1, (3, 1), 1, padding=(1, 0)))

    def forward(self, x):
        fmap = []

        # 1d to 2d
        b, c, t = x.shape
        if t % self.period != 0:  # pad first
            n_pad = self.period - (t % self.period)
            x = F.pad(x, (0, n_pad), "reflect")
            t = t + n_pad
        x = x.view(b, c, t // self.period, self.period)

        for layer in self.convs:
            x = layer(x)
            x = self.config.model.activation_function(x)
            fmap.append(x)
        x = self.conv_post(x)
        fmap.append(x)
        x = torch.flatten(x, 1, -1)

        return x, fmap


class MultiPeriodDiscriminator(torch.nn.Module):
    def __init__(self, config):
        super(MultiPeriodDiscriminator, self).__init__()
        mpd_layers = [DiscriminatorP(n, config) for n in config.model.mpd_layers]
        self.discriminators = nn.ModuleList(mpd_layers)

    def forward_interpolates(self, interp):
        y_ds = []
        f_maps = []
        for d in self.discriminators:
            y_d, f_map = d(interp)
            y_ds.append(y_d)
            f_maps.append(f_map)
        return y_ds, f_maps

    def forward(self, y, y_hat):
        y_d_rs = []
        y_d_gs = []
        fmap_rs = []
        fmap_gs = []
        for d in self.discriminators:
            y_d_r, fmap_r = d(y)
            y_d_g, fmap_g = d(y_hat)
            y_d_rs.append(y_d_r)
            fmap_rs.append(fmap_r)
            y_d_gs.append(y_d_g)
            fmap_gs.append(fmap_g)

        return y_d_rs, y_d_gs, fmap_rs, fmap_gs


class DiscriminatorS(torch.nn.Module):
    def __init__(self, config, use_spectral_norm=False):
        super(DiscriminatorS, self).__init__()
        self.config = config
        norm_f = weight_norm if use_spectral_norm is False else spectral_norm
        self.convs = nn.ModuleList(
            [
                norm_f(Conv1d(1, 128, 15, 1, padding=7)),
                norm_f(Conv1d(128, 128, 41, 2, groups=4, padding=20)),
                norm_f(Conv1d(128, 256, 41, 2, groups=16, padding=20)),
                norm_f(Conv1d(256, 512, 41, 4, groups=16, padding=20)),
                norm_f(Conv1d(512, 1024, 41, 4, groups=16, padding=20)),
                norm_f(Conv1d(1024, 1024, 41, 1, groups=16, padding=20)),
                norm_f(Conv1d(1024, 1024, 5, 1, padding=2)),
            ]
        )
        self.conv_post = norm_f(Conv1d(1024, 1, 3, 1, padding=1))

    def forward(self, x):
        fmap = []
        for layer in self.convs:
            x = layer(x)
            x = self.config.model.activation_function(x)
            fmap.append(x)
        x = self.conv_post(x)
        fmap.append(x)
        x = torch.flatten(x, 1, -1)

        return x, fmap


class MultiScaleDiscriminator(torch.nn.Module):
    def __init__(self, config: VocoderConfig):
        super(MultiScaleDiscriminator, self).__init__()
        msd_layers = [
            DiscriminatorS(config, use_spectral_norm=i == 0)
            for i in range(config.model.msd_layers)
        ]
        self.discriminators = nn.ModuleList(msd_layers)
        self.meanpools = nn.ModuleList(
            [AvgPool1d(4, 2, padding=2) for _ in range(config.model.msd_layers - 1)]
        )

    def forward_interpolates(self, interp):
        y_ds = []
        f_maps = []
        for i, d in enumerate(self.discriminators):
            if i != 0:
                interp = self.meanpools[i - 1](interp)
            y_d, f_map = d(interp)
            y_ds.append(y_d)
            f_maps.append(f_map)
        return y_ds, f_maps

    def forward(self, y, y_hat):
        y_d_rs = []
        y_d_gs = []
        fmap_rs = []
        fmap_gs = []
        for i, d in enumerate(self.discriminators):
            if i != 0:
                y = self.meanpools[i - 1](y)
                y_hat = self.meanpools[i - 1](y_hat)
            y_d_r, fmap_r = d(y)
            y_d_g, fmap_g = d(y_hat)
            y_d_rs.append(y_d_r)
            fmap_rs.append(fmap_r)
            y_d_gs.append(y_d_g)
            fmap_gs.append(fmap_g)

        return y_d_rs, y_d_gs, fmap_rs, fmap_gs


class HiFiGAN(pl.LightningModule):
    def __init__(self, config: dict | VocoderConfig):
        super().__init__()
        # Required by PyTorch Lightning for our manual optimization
        self.automatic_optimization = False
        # Because we serialize the configurations when saving checkpoints,
        # sometimes what is passed is actually just a dict.
        if not isinstance(config, VocoderConfig):
            config = VocoderConfig(**config)
        self.config = config
        self.mpd = MultiPeriodDiscriminator(config)
        self.msd = MultiScaleDiscriminator(config)
        self.generator = Generator(config)
        self.save_hyperparameters()  # TODO: ignore=['specific keys'] - I should ignore some unnecessary/problem values
        self.update_config_settings()
        if self.config.model.istft_layer:
            self.inverse_spectral_transform = get_spectral_transform(
                "istft",
                self.generator.post_n_fft,
                self.generator.post_n_fft,
                self.generator.post_n_fft // 4,
            )
        # TODO: figure out multiple nodes/gpus: https://pytorch-lightning.readthedocs.io/en/1.4.0/advanced/multi_gpu.html

    def update_config_settings(self):
        # batch_size is declared explicitly so that auto_scale_batch_size works:
        # https://pytorch-lightning.readthedocs.io/en/stable/advanced/training_tricks.html
        if self.config.training.finetune:
            for disc in self.mpd.discriminators:
                for layer in disc.convs[:3]:
                    for p in layer.parameters():
                        p.requires_grad = False
            for disc in self.msd.discriminators:
                for layer in disc.convs[:4]:
                    for p in layer.parameters():
                        p.requires_grad = False
        self.batch_size = self.config.training.batch_size
        self.audio_config = self.config.preprocessing.audio
        self.sampling_rate_change = (
            self.audio_config.output_sampling_rate
            // self.audio_config.input_sampling_rate
        )
        self.use_wgan = self.config.training.gan_type is HiFiGANTrainTypes.wgan
        # We don't have to set the fft size and hop/window lengths as hyperparameters here, because we can just multiply by the upsampling rate
        self.spectral_transform = get_spectral_transform(
            self.audio_config.spec_type,
            self.audio_config.n_fft * self.sampling_rate_change,
            self.audio_config.fft_window_size * self.sampling_rate_change,
            self.audio_config.fft_hop_size * self.sampling_rate_change,
            f_min=self.audio_config.f_min,
            f_max=self.audio_config.f_max,
            sample_rate=self.audio_config.output_sampling_rate,
            n_mels=self.audio_config.n_mels,
        )

    def forward(self, x):
        return self.generator(x)

    def on_load_checkpoint(self, checkpoint):
        """Deserialize the checkpoint hyperparameters.
        Note, this shouldn't fail on different versions of pydantic anymore,
        but it will fail on breaking changes to the config. We should catch those exceptions
        and handle them appropriately."""
        self.config = VocoderConfig(**checkpoint["hyper_parameters"]["config"])

    def on_save_checkpoint(self, checkpoint):
        """Serialize the checkpoint hyperparameters"""
        checkpoint["hyper_parameters"]["config"] = self.config.model_checkpoint_dump()

    def configure_optimizers(self):
        generator_params = self.generator.parameters()
        if self.config.training.finetune:
            msd_params = filter(lambda p: p.requires_grad, self.msd.parameters())
            mpd_params = filter(lambda p: p.requires_grad, self.mpd.parameters())
        else:
            msd_params = self.msd.parameters()
            mpd_params = self.mpd.parameters()
        if self.config.training.optimizer.name == "adamw":
            optim_g = torch.optim.AdamW(
                generator_params,
                self.config.training.optimizer.learning_rate,
                betas=[
                    self.config.training.optimizer.betas[0],
                    self.config.training.optimizer.betas[1],
                ],
            )
            optim_d = torch.optim.AdamW(
                itertools.chain(msd_params, mpd_params),
                self.config.training.optimizer.learning_rate,
                betas=[
                    self.config.training.optimizer.betas[0],
                    self.config.training.optimizer.betas[1],
                ],
            )
        elif self.config.training.optimizer.name == "rms":
            optim_g = torch.optim.RMSprop(
                generator_params,
                lr=self.config.training.optimizer.learning_rate,
            )
            optim_d = torch.optim.RMSprop(
                itertools.chain(msd_params, mpd_params),
                lr=self.config.training.optimizer.learning_rate,
            )
        if self.use_wgan:
            return (
                {"optimizer": optim_g, "frequency": 1},
                {"optimizer": optim_d, "frequency": 5},
            )
        else:
            scheduler_g = torch.optim.lr_scheduler.ExponentialLR(
                optim_g,
                gamma=0.999,  # TODO: parametrize this
            )
            scheduler_d = torch.optim.lr_scheduler.ExponentialLR(optim_d, gamma=0.999)
            return [optim_g, optim_d], [scheduler_g, scheduler_d]

    def feature_loss(self, fmap_r, fmap_g):
        loss = 0
        for dr, dg in zip(fmap_r, fmap_g):
            for rl, gl in zip(dr, dg):
                loss += torch.mean(torch.abs(rl - gl))

        return loss * 2

    def discriminator_loss(
        self,
        disc_real_outputs,
        disc_generated_outputs,
    ):
        loss = 0
        r_losses = []
        g_losses = []
        for dr, dg in zip(disc_real_outputs, disc_generated_outputs):
            r_loss = torch.mean((1 - dr) ** 2)
            g_loss = torch.mean(dg**2)
            loss += r_loss + g_loss
            r_losses.append(r_loss.item())
            g_losses.append(g_loss.item())

        return loss, r_losses, g_losses

    def generator_loss(self, disc_outputs):
        g_loss = 0
        gen_losses = []
        for dg in disc_outputs:
            loss = torch.mean((1 - dg) ** 2)
            gen_losses.append(loss)
            g_loss += loss

        return (g_loss, gen_losses)

    def training_step(self, batch, batch_idx):
        x, y, _, y_mel = batch
        y = y.unsqueeze(1)
        # x.size() & y_mel.size() = [batch_size, n_mels=80, n_frames=32]
        # y.size() = [batch_size, segment_size=8192]
        optim_g, optim_d = self.optimizers()
        scheduler_g, scheduler_d = self.lr_schedulers()
        # generate waveform
        if self.config.model.istft_layer:
            mag, phase = self(x)
            generated_wav = self.inverse_spectral_transform(
                mag * torch.exp(phase * 1j)
            ).unsqueeze(-2)
        else:
            generated_wav = self(x)

        # create mel
        generated_mel_spec = dynamic_range_compression_torch(
            self.spectral_transform(generated_wav).squeeze(1)[:, :, 1:]
        )
        # train discriminators
        if self.global_step >= self.config.training.generator_warmup_steps:
            optim_d.zero_grad()
            # MPD
            y_df_hat_r, y_df_hat_g, _, _ = self.mpd(y, generated_wav.detach())
            loss_disc_f, _, _ = self.discriminator_loss(y_df_hat_r, y_df_hat_g)
            self.log("training/disc/mpd_loss", loss_disc_f, prog_bar=False)
            # MSD
            y_ds_hat_r, y_ds_hat_g, _, _ = self.msd(y, generated_wav.detach())
            loss_disc_s, _, _ = self.discriminator_loss(y_ds_hat_r, y_ds_hat_g)
            self.log("training/disc/msd_loss", loss_disc_s, prog_bar=False)
            # WGAN
            if self.use_wgan:
                for p in self.msd.parameters():
                    p.data.clamp_(
                        -self.config.training.wgan_clip_value,
                        self.config.training.wgan_clip_value,
                    )
                for p in self.mpd.parameters():
                    p.data.clamp_(
                        -self.config.training.wgan_clip_value,
                        self.config.training.wgan_clip_value,
                    )
            # calculate loss
            disc_loss_total = loss_disc_s + loss_disc_f
            # manual optimization because Pytorch Lightning 2.0+ doesn't handle automatic optimization for multiple optimizers
            # use .backward for now, but maybe switch to self.manual_backward() in the future: https://github.com/Lightning-AI/lightning/issues/18740
            # self.manual_backward(disc_loss_total)
            disc_loss_total.backward()
            # clip gradients
            self.clip_gradients(
                optim_d, gradient_clip_val=1.0, gradient_clip_algorithm="norm"
            )
            optim_d.step()
            # step in the scheduler every epoch
            if self.trainer.is_last_batch:
                scheduler_d.step()
            # log discriminator loss
            self.log("training/disc/d_loss_total", disc_loss_total, prog_bar=False)

        # train generator
        optim_g.zero_grad()
        # calculate loss
        loss_mel = F.l1_loss(y_mel, generated_mel_spec) * 45
        if self.global_step >= self.config.training.generator_warmup_steps:
            _, y_df_hat_g, fmap_f_r, fmap_f_g = self.mpd(y, generated_wav)
            _, y_ds_hat_g, fmap_s_r, fmap_s_g = self.msd(y, generated_wav)
            loss_fm_f = self.feature_loss(fmap_f_r, fmap_f_g)
            loss_fm_s = self.feature_loss(fmap_s_r, fmap_s_g)
            # loss_gen_f = -torch.mean(y_df_hat_g)
            # loss_gen_s = -torch.mean(y_ds_hat_g)
            loss_gen_f, _ = self.generator_loss(y_df_hat_g)
            loss_gen_s, _ = self.generator_loss(y_ds_hat_g)
            self.log("training/gen/loss_fmap_f", loss_fm_f, prog_bar=False)
            self.log("training/gen/loss_fmap_s", loss_fm_s, prog_bar=False)
            self.log("training/gen/loss_gen_f", loss_gen_f, prog_bar=False)
            self.log("training/gen/loss_gen_s", loss_gen_s, prog_bar=False)
            gen_loss_total = loss_gen_s + loss_gen_f + loss_fm_s + loss_fm_f + loss_mel
        else:
            gen_loss_total = loss_mel
        # manual optimization because Pytorch Lightning 2.0+ doesn't handle automatic optimization for multiple optimizers
        # use .backward for now, but maybe switch to self.manual_backward() in the future: https://github.com/Lightning-AI/lightning/issues/18740
        # self.manual_backward(gen_loss_total)
        gen_loss_total.backward()
        # clip gradients
        self.clip_gradients(
            optim_d, gradient_clip_val=1.0, gradient_clip_algorithm="norm"
        )
        optim_g.step()
        # step in the scheduler every epoch
        if self.trainer.is_last_batch:
            scheduler_g.step()
        # log generator loss
        self.log("training/gen/gen_loss_total", gen_loss_total, prog_bar=False)
        self.log("training/gen/mel_spec_error", loss_mel / 45, prog_bar=False)

    def validation_step(self, batch, batch_idx):
        x, y, bn, y_mel = batch
        current_step = (
            self.global_step // 2
        )  # because self.global_step counts gen/disc steps
        # generate waveform
        if self.config.model.istft_layer:
            mag, phase = self(x)
            generated_wav = self.inverse_spectral_transform(
                mag * torch.exp(phase * 1j)
            ).unsqueeze(-2)
        else:
            generated_wav = self(x)
        # create mel
        generated_mel_spec = dynamic_range_compression_torch(
            self.spectral_transform(generated_wav).squeeze(1)[:, :, 1:]
        )
        # Since we are not using fixed-size segments, sometimes the prediction is off by one frame when doing super resolution/upsampling
        val_err_tot = F.l1_loss(y_mel, generated_mel_spec[:, :, : y_mel.size(2)]).item()
        # # Below is taken from HiFiGAN
        if self.global_step == 0:
            # Log ground truth audio and spec
            self.logger.experiment.add_audio(
                f"gt/y_{bn[0]}",
                y[0],
                current_step,
                self.audio_config.output_sampling_rate,
            )
            self.logger.experiment.add_figure(
                f"gt/y_spec_{bn[0]}",
                plot_spectrogram(x[0].cpu().numpy()),
                current_step,
            )
        #
        if batch_idx == 0:
            self.logger.experiment.add_audio(
                f"generated/y_hat_{bn[0]}",
                generated_wav[0],
                current_step,
                self.audio_config.output_sampling_rate,
            )

            y_hat_spec = dynamic_range_compression_torch(
                self.spectral_transform(generated_wav[0]).squeeze(1)[:, :, 1:]
            )
            self.logger.experiment.add_figure(
                f"generated/y_hat_spec_{bn[0]}",
                plot_spectrogram(y_hat_spec.squeeze(0).cpu().numpy()),
                current_step,
            )

        self.log(
            "validation/mel_spec_error", val_err_tot, prog_bar=True, sync_dist=True
        )



from enum import Enum
from pathlib import Path
from typing import Annotated, List, Optional, Union

from annotated_types import Ge, Le
from pydantic import Field, FilePath, ValidationInfo, field_validator, model_validator

from everyvoice.config.shared_types import ConfigModel, PartialLoadConfig, init_context
from everyvoice.config.utils import (
    PossiblyRelativePath,
    PossiblyRelativePathMustExist,
    PossiblySerializedCallable,
    load_partials,
)
from everyvoice.utils import (
    generic_psv_filelist_reader,
    load_config_from_json_or_yaml_path,
)


class AudioSpecTypeEnum(str, Enum):
    mel = "mel"  # TorchAudio implementation
    mel_librosa = "mel-librosa"  # Librosa implementation
    linear = "linear"  # TorchAudio Linear Spectrogram
    raw = "raw"  # TorchAudio Complex Spectrogram


class AudioConfig(ConfigModel):
    min_audio_length: float = Field(
        0.4,
        description="The minimum length of an audio sample in seconds. Audio shorter than this will be ignored during preprocessing.",
    )
    max_audio_length: float = Field(
        11.0,
        description="The maximum length of an audio sample in seconds. Audio longer than this will be ignored during preprocessing. Increasing the max_audio_length will result in larger memory usage. If you are running out of memory, consider lowering the max_audio_length.",
    )
    max_wav_value: float = Field(
        32767.0,
        description="Advanced. The maximum value allowed to be in your wav files. For 16-bit audio, this should be (2**16)/2 - 1.",
    )
    input_sampling_rate: int = Field(
        22050,
        description="The sampling rate describes the number of samples per second of audio. The 'input_sampling_rate' is with respect to your vocoder, or spec-to-wav model. This means that the spectrograms predicted by your text-to-spec model will also be calculated from audio at this sampling rate. If you change this value, your audio will automatically be re-sampled during preprocessing.",
    )
    output_sampling_rate: int = Field(
        22050,
        description="Advanced. The sampling rate describes the number of samples per second of audio. The 'output_sampling_rate' is with respect to your vocoder, or spec-to-wav model. This means that the wav files generated by your vocoder or spec-to-wav model will be at this sampling rate. If you change this value, you will also need to change the upsample rates in your vocoder. Your audio will automatically be re-sampled during preprocessing.",
    )
    alignment_sampling_rate: int = Field(
        22050,
        description="Advanced. The sampling rate describes the number of samples per second of audio. The 'alignment_sampling_rate' describes the sampling rate used when training an alignment model. If you change this value, your audio will automatically be re-sampled during preprocessing.",
    )
    target_bit_depth: int = Field(
        16,
        description="Advanced. This is the bit depth of each sample in your audio files.",
    )
    n_fft: int = Field(
        1024,
        title="FFT Size",
        description="Advanced. This is the number of bins used by the Fast Fourier Transform (FFT).",
    )
    fft_window_size: int = Field(
        1024,
        title="FFT Window Size",
        description="Advanced. This is the window size used by the Fast Fourier Transform (FFT).",
    )
    fft_hop_size: int = Field(
        256,
        title="FFT Hop Size",
        description="Advanced. This is the hop size for calculating the Short-Time Fourier Transform (STFT) which calculates a sequence of spectrograms from a single audio file. Another way of putting it is that the hop size is equal to the amount of non-intersecting samples from the audio in each spectrogram.",
    )
    f_min: int = Field(
        0,
        title="Minimum Frequency",
        description="Advanced. This is the minimum frequency for the lowest frequency bin when calculating the spectrogram.",
    )
    f_max: int = Field(
        8000,
        title="Maximum Frequency",
        description="Advanced. This is the maximum frequency for the highest frequency bin when calculating the spectrogram.",
    )
    n_mels: int = Field(
        80,
        title="Number of Mel bins",
        description="Advanced. This is the number of filters in the Mel-scale spaced filterbank.",
    )
    spec_type: Union[AudioSpecTypeEnum, str] = Field(
        AudioSpecTypeEnum.mel_librosa.value,
        description="Advanced. Defines how to calculate the spectrogram. 'mel' uses the TorchAudio implementation for a Mel spectrogram. 'mel-librosa' uses Librosa's implementation. 'linear' calculates a non-Mel linear spectrogram and 'raw' calculates a complex-valued spectrogram. 'linear' and 'raw' are not currently supported by EveryVoice. We recommend using 'mel-librosa'.",
    )
    vocoder_segment_size: int = Field(
        8192,
        description="Advanced. The vocoder, or spec-to-wav model is trained by sampling random fixed-size sections of the audio. This value specifies the number of samples in those sections.",
    )


class Dataset(PartialLoadConfig):
    label: str = Field("YourDataSet", description="A label for the source of data")
    permissions_obtained: bool = Field(
        False,
        description="An attestation that permission has been obtained to use this data. You may not use EveryVoice to build a TTS system with data that you do not have permission to use and there are serious possible consequences for doing so. Finding data online does not constitute permission. The speaker should be aware and consent to their data being used in this way.",
        validate_default=True,
    )
    data_dir: PossiblyRelativePath = Field(
        Path("/please/create/a/path/to/your/dataset/data"),
        description="The path to the directory with your audio files.",
    )
    filelist: PossiblyRelativePath = Field(
        Path("/please/create/a/path/to/your/dataset/filelist"),
        description="The path to your dataset's filelist.",
    )
    filelist_loader: PossiblySerializedCallable = Field(
        generic_psv_filelist_reader,
        description="Advanced. The file-loader function to use to load your dataset's filelist.",
    )
    sox_effects: list = Field(
        [["channels", "1"]],
        description="Advanced. A list of SoX effects to apply to your audio prior to preprocessing. Run python -c 'import torchaudio; print(torchaudio.sox_effects.effect_names())' to see a list of supported effects.",
    )

    @field_validator("permissions_obtained")
    def check_permissions(cls, permissions_obtained: bool) -> bool:
        if not permissions_obtained:
            raise ValueError(
                "You are trying to run a model that does not have permission for the data it is using. Please confirm you have permission to use this data and edit your configuration file accordingly."
            )
        return permissions_obtained


class PreprocessingConfig(PartialLoadConfig):
    dataset: str = Field("YourDataSet", description="The name of the dataset.")
    train_split: Annotated[float, Ge(0.0), Le(1.0)] = Field(
        0.9,
        description="The amount of the dataset to use for training. The rest will be used as validation. Hold some of the validation set out for a test set if you are performing experiments.",
    )
    dataset_split_seed: int = Field(
        1234,
        description="The seed to use when splitting the dataset into train and validation sets.",
    )
    save_dir: PossiblyRelativePathMustExist = Field(
        Path("./preprocessed/YourDataSet"),
        description="The directory to save preprocessed files to.",
    )
    audio: AudioConfig = Field(
        default_factory=AudioConfig,
        description="Configuration settings for audio.",
    )
    path_to_audio_config_file: Optional[FilePath] = Field(
        None, description="The path to an audio configuration file."
    )
    source_data: List[Dataset] = Field(
        default_factory=lambda: [
            Dataset(permissions_obtained=True)
        ],  # The default factory doesn't actually point to any data, so we can treat it as having permissions obtained
        description="A list of datasets.",
    )

    @model_validator(mode="before")  # type: ignore
    def load_partials(self, info: ValidationInfo):
        config_path = (
            info.context.get("config_path", None) if info.context is not None else None
        )
        return load_partials(
            self,  # type: ignore
            ("audio",),
            config_path=config_path,
        )

    @staticmethod
    def load_config_from_path(path: Path) -> "PreprocessingConfig":
        """Load a config from a path"""
        config = load_config_from_json_or_yaml_path(path)
        with init_context({"config_path": path}):
            config = PreprocessingConfig(**config)
        return config



from multiprocessing import managers
from pathlib import Path

import torch
import torchaudio


def save_tensor(tensor: torch.Tensor, path: str | Path):
    """Create hierarchy before saving a tensor."""
    path = Path(path)
    if not path.parent.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
    torch.save(tensor, path)


def save_wav(
    audio: torch.Tensor, path: str | Path, sr: int | None, bits_per_sample: int
):
    """Create hierarchy before saving a wav."""
    path = Path(path)
    if not path.parent.exists():
        path.parent.mkdir(parents=True, exist_ok=True)
    torchaudio.save(
        str(path),
        audio,
        sr,
        encoding="PCM_S",
        bits_per_sample=bits_per_sample,
    )


class Scaler:
    def __init__(self):
        self._data = []
        self._tensor_data = None
        self.min = None
        self.max = None
        self.std = None
        self.mean = None
        self.norm_min = None
        self.norm_max = None

    def __len__(self):
        return len(self.data)

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, value):
        raise ValueError(
            f"Sorry, you tried to change the data to {value} but it cannot be changed directly. Either Scaler.append(data), or Scaler.clear_data()"
        )

    def append(self, value):
        self._data += value

    def clear_data(self):
        """Clear data"""
        self.__init__()

    def normalize(self, data):
        """Remove mean and normalize to unit variance"""
        return (data - self.mean) / self.std

    def denormalize(self, data):
        """Get de-normalized value"""
        return (data * self.std) + self.mean

    def calculate_stats(self):
        if not len(self):
            return
        if self._tensor_data is None:
            self._tensor_data = torch.cat(self.data)
        non_nan_data = self._tensor_data[~torch.isnan(self._tensor_data)]
        self.min = torch.min(non_nan_data)
        self.max = torch.max(non_nan_data)
        self.mean = torch.nanmean(self._tensor_data)
        self.std = torch.std(non_nan_data)
        self.norm_max = self.normalize(self.max)
        self.norm_min = self.normalize(self.min)
        return {
            "sample_size": len(self),
            "norm_min": float(self.norm_min),
            "norm_max": float(self.norm_max),
            "min": float(self.min),
            "max": float(self.max),
            "mean": float(self.mean),
            "std": float(self.std),
        }


class Counters:
    def __init__(self, manager: managers.SyncManager):
        self._lock = manager.Lock()
        self._processed_files = manager.Value("l", 0)
        self._previously_processed_files = manager.Value("l", 0)
        self._duration = manager.Value("d", 0)
        self._nans = manager.Value("l", 0)
        self._audio_empty = manager.Value("l", 0)
        self._audio_too_long = manager.Value("l", 0)
        self._audio_too_short = manager.Value("l", 0)
        self._skipped_processes = manager.Value("l", 0)
        self._missing_files = manager.Value("l", 0)

    def increment(self, counter: str, increment: int | float = 1):
        with self._lock:
            self.__getattribute__("_" + counter).value += increment

    def value(self, counter):
        with self._lock:
            return self.__getattribute__("_" + counter).value



"""
This file is for light-weight type definitions with no dependencies.
This should load in milliseconds.
Expensive type definitions, e.g., requiring pydantic, belong in type_definitions_heavy.py.
"""

from enum import Enum


class SynthesizeOutputFormats(str, Enum):
    """Valid output formats for synthesize"""

    wav = "wav"
    spec = "spec"
    textgrid = "textgrid"



from tqdm import tqdm

from everyvoice.config.type_definitions import TargetTrainingTextRepresentationLevel
from everyvoice.dataloader import BaseDataModule
from everyvoice.dataloader.imbalanced_sampler import ImbalancedDatasetSampler
from everyvoice.model.aligner.config import AlignerConfig
from everyvoice.model.e2e.config import EveryVoiceConfig
from everyvoice.model.feature_prediction.config import FeaturePredictionConfig
from everyvoice.model.vocoder.config import VocoderConfig
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.config import (
    HiFiGANTrainingConfig,
    PreprocessingConfig,
)
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.dataset import (
    HiFiGANDataModule,
    SpecDataset,
)
from everyvoice.tests.basic_test_case import BasicTestCase
from everyvoice.tests.preprocessed_audio_fixture import PreprocessedAudioFixture
from everyvoice.utils import filter_dataset_based_on_target_text_representation_level


class DataLoaderTest(PreprocessedAudioFixture, BasicTestCase):
    """Basic test for dataloaders"""

    def setUp(self) -> None:
        super().setUp()

        self.config = EveryVoiceConfig(
            contact=BasicTestCase.contact,
            aligner=AlignerConfig(contact=BasicTestCase.contact),
            feature_prediction=FeaturePredictionConfig(contact=BasicTestCase.contact),
            vocoder=VocoderConfig(
                contact=BasicTestCase.contact,
                training=HiFiGANTrainingConfig(
                    training_filelist=PreprocessedAudioFixture.lj_preprocessed
                    / "preprocessed_filelist.psv",
                    validation_filelist=PreprocessedAudioFixture.lj_preprocessed
                    / "validation_preprocessed_filelist.psv",
                ),
                preprocessing=PreprocessingConfig(
                    save_dir=PreprocessedAudioFixture.lj_preprocessed,
                ),
            ),
        )

    def test_base_data_loader(self):
        bdm = BaseDataModule(self.config.aligner)
        with self.assertRaises(NotImplementedError):
            bdm.load_dataset()

    def test_spec_dataset(self):
        dataset = SpecDataset(
            self.config.vocoder.training.filelist_loader(
                self.config.vocoder.training.training_filelist
            ),
            self.config.vocoder,
            use_segments=True,
        )
        for sample in tqdm(dataset):
            spec, audio, basename, spec_from_audio = sample
            self.assertTrue(isinstance(basename, str))
            self.assertEqual(spec.size(), spec_from_audio.size())
            self.assertEqual(
                spec.size(0), self.config.vocoder.preprocessing.audio.n_mels
            )
            self.assertEqual(
                spec.size(1),
                self.config.vocoder.preprocessing.audio.vocoder_segment_size
                / (
                    self.config.vocoder.preprocessing.audio.fft_hop_size
                    * (
                        self.config.vocoder.preprocessing.audio.output_sampling_rate
                        // self.config.vocoder.preprocessing.audio.input_sampling_rate
                    )
                ),
            )

    def test_hifi_data_loader(self):
        hfgdm = HiFiGANDataModule(self.config.vocoder)
        hfgdm.load_dataset()
        self.assertEqual(len(hfgdm.train_dataset), 5)

    def test_filter_dataset(self):
        train_dataset = [{"character_tokens": "b", "phone_tokens": ""}] * 4
        with self.assertRaises(SystemExit) as cm:
            filter_dataset_based_on_target_text_representation_level(
                TargetTrainingTextRepresentationLevel.characters,
                train_dataset,
                "training",
                6,
            )
        self.assertEqual(cm.exception.code, 1)
        with self.assertRaises(SystemExit) as cm:
            filter_dataset_based_on_target_text_representation_level(
                TargetTrainingTextRepresentationLevel.ipa_phones,
                train_dataset,
                "training",
                4,
            )
        self.assertEqual(cm.exception.code, 1)
        train_ds = filter_dataset_based_on_target_text_representation_level(
            TargetTrainingTextRepresentationLevel.characters,
            train_dataset,
            "training",
            4,
        )
        val_ds = filter_dataset_based_on_target_text_representation_level(
            TargetTrainingTextRepresentationLevel.characters,
            train_dataset,
            "validation",
            4,
        )
        self.assertEqual(len(train_ds), 4)
        self.assertEqual(len(val_ds), 4)

    def test_hifi_ft_data_loader(self):
        """TODO: can't make this test until I generate some synthesized samples"""
        pass

    def test_feature_prediction_data_loader(self):
        # TODO: once feature prediction is done
        pass

    def test_e2e_data_module(self):
        # TODO: once e2e is done
        pass

    def test_imbalanced_sampler(self):
        dataset = SpecDataset(
            self.config.vocoder.training.filelist_loader(
                self.config.vocoder.training.training_filelist
            ),
            self.config.vocoder,
            use_segments=True,
        )
        sampler = ImbalancedDatasetSampler(dataset)
        sample = list(sampler)
        self.assertEqual(len(sample), 5)



from functools import partial
from pathlib import Path

import numpy as np
import torch
from everyvoice.config.type_definitions import (
    DatasetTextRepresentation,
    TargetTrainingTextRepresentationLevel,
)
from everyvoice.dataloader import BaseDataModule
from everyvoice.preprocessor import Preprocessor
from everyvoice.text.lookups import LookupTable, lookuptables_from_config
from everyvoice.text.text_processor import TextProcessor
from everyvoice.utils import (
    _flatten,
    check_dataset_size,
    filter_dataset_based_on_target_text_representation_level,
)
from torch.nn.utils.rnn import pad_sequence
from torch.utils.data import Dataset

from .config import FastSpeech2Config


class FastSpeechDataset(Dataset):
    """
    To debug, set num_workers=0 and batch_size=1
    """

    def __init__(
        self,
        dataset,
        config: FastSpeech2Config,
        lang2id: LookupTable,
        speaker2id: LookupTable,
        teacher_forcing=False,
        inference=False,
    ):
        self.dataset = dataset
        self.config = config
        self.sep = "--"
        self.text_processor = TextProcessor(config.text)
        self.preprocessed_dir = Path(self.config.preprocessing.save_dir)
        self.sampling_rate = self.config.preprocessing.audio.input_sampling_rate
        self.teacher_forcing = teacher_forcing
        self.inference = inference
        self.lang2id = lang2id
        self.speaker2id = speaker2id

    def _load_file(self, bn, spk, lang, dir, fn):
        return torch.load(
            self.preprocessed_dir / dir / self.sep.join([bn, spk, lang, fn])
        )

    def __getitem__(self, index):
        """
        Returns dict with keys: {
            "mel"
            "duration"
            "duration_control"
            "pfs"
            "text"
            "raw_text"
            "basename"
            "speaker"
            "speaker_id"
            "language"
            "language_id"
            "label"
            "energy"
            "pitch"
        }
        """
        item = self.dataset[index]
        speaker = "default" if "speaker" not in item else item["speaker"]
        language = "default" if "language" not in item else item["language"]
        duration_control = (
            1.0 if "duration_control" not in item else item["duration_control"]
        )
        speaker_id = self.speaker2id[speaker]
        language_id = self.lang2id[language]
        basename = item["basename"]
        if self.inference:
            # TODO: we shouldn't calculate all possible representations at synthesis time,
            #       despite that's what we do at preprocessing time.
            character_tokens, phone_tokens, _ = Preprocessor.process_text(
                item,
                text_processor=self.text_processor,
                use_pfs=False,
                encode_as_string=True,
            )
            item["character_tokens"] = character_tokens
            item["phone_tokens"] = phone_tokens
        if self.teacher_forcing or not self.inference:
            mel = self._load_file(
                basename,
                speaker,
                language,
                "spec",
                f"spec-{self.sampling_rate}-{self.config.preprocessing.audio.spec_type}.pt",
            ).transpose(
                0, 1
            )  # [mel_bins, frames] -> [frames, mel_bins]
        else:
            mel = None
        if (
            self.teacher_forcing or not self.inference
        ) and self.config.model.learn_alignment:
            match self.config.model.target_text_representation_level:
                case TargetTrainingTextRepresentationLevel.characters:
                    duration = self._load_file(
                        basename,
                        speaker,
                        language,
                        "attn",
                        f"{DatasetTextRepresentation.characters.value}-attn-prior.pt",
                    )
                case TargetTrainingTextRepresentationLevel.ipa_phones | TargetTrainingTextRepresentationLevel.phonological_features:
                    duration = self._load_file(
                        basename,
                        speaker,
                        language,
                        "attn",
                        f"{DatasetTextRepresentation.ipa_phones.value}-attn-prior.pt",
                    )
                case _:
                    raise NotImplementedError(
                        f"{self.config.model.target_text_representation_level} have not yet been implemented."
                    )
        elif self.teacher_forcing or not self.inference:
            duration = self._load_file(
                basename, speaker, language, "duration", "duration.pt"
            )
        else:
            duration = None
        match self.config.model.target_text_representation_level:
            case TargetTrainingTextRepresentationLevel.characters:
                text = torch.IntTensor(
                    self.text_processor.encode_escaped_string_sequence(
                        item["character_tokens"]
                    )
                )
            case TargetTrainingTextRepresentationLevel.ipa_phones | TargetTrainingTextRepresentationLevel.phonological_features:
                text = torch.IntTensor(
                    self.text_processor.encode_escaped_string_sequence(
                        item["phone_tokens"]
                    )
                )
            case _:
                raise NotImplementedError(
                    f"{self.config.model.target_text_representation_level} have not yet been implemented."
                )

        if TargetTrainingTextRepresentationLevel.characters.value in item:
            raw_text = item[TargetTrainingTextRepresentationLevel.characters.value]
        else:
            raw_text = item.get(
                TargetTrainingTextRepresentationLevel.ipa_phones.value, "text"
            )
        pfs = None
        if (
            self.config.model.target_text_representation_level
            == TargetTrainingTextRepresentationLevel.phonological_features
        ):
            pfs = self._load_file(basename, speaker, language, "pfs", "pfs.pt")
        if not self.inference:
            energy = self._load_file(basename, speaker, language, "energy", "energy.pt")
            pitch = self._load_file(basename, speaker, language, "pitch", "pitch.pt")
        else:
            energy = None
            pitch = None

        return {
            "mel": mel,
            "duration": duration,
            "duration_control": duration_control,
            "pfs": pfs,
            "text": text,
            "raw_text": raw_text,
            "basename": basename,
            "speaker": speaker,
            "speaker_id": speaker_id,
            "language": language,
            "language_id": language_id,
            # "label": item.get("label", "default"), # TODO: determine proper label
            "energy": energy,
            "pitch": pitch,
        }

    def __len__(self):
        return len(self.dataset)

    def get_labels(self):
        return [x["label"] for x in self.dataset]


class FastSpeech2DataModule(BaseDataModule):
    def __init__(
        self,
        config: FastSpeech2Config,
        inference=False,
        teacher_forcing=False,
        inference_output_dir=Path("synthesis_output"),
    ):
        super().__init__(config=config, inference_output_dir=inference_output_dir)
        self.inference = inference
        self.prepared = False
        self.teacher_forcing = teacher_forcing
        self.collate_fn = partial(
            self.collate_method, learn_alignment=config.model.learn_alignment
        )
        self.use_weighted_sampler = config.training.use_weighted_sampler
        self.batch_size = config.training.batch_size
        if not inference:
            self.load_dataset()
            self.dataset_length = len(self.train_dataset) + len(self.val_dataset)
            self.lang2id, self.speaker2id = lookuptables_from_config(config)

    @staticmethod
    def collate_method(data, learn_alignment=True):
        data = [_flatten(x) for x in data]
        data = {k: [dic[k] for dic in data] for k in data[0]}
        text_lens = torch.IntTensor([text.size(0) for text in data["text"]])
        max_text = max(text_lens)
        if data["mel"][0] is not None:
            mel_lens = torch.IntTensor([mel.size(0) for mel in data["mel"]])
            max_mel = max(mel_lens)
        else:
            mel_lens = None
            max_mel = 1_000_000

        for key in data:
            if isinstance(data[key][0], np.ndarray):
                data[key] = [torch.tensor(x) for x in data[key]]
            if torch.is_tensor(data[key][0]):
                if key == "duration" and learn_alignment:
                    # in this case we need to pad both the src and target dimensions
                    dur_padded = torch.zeros(len(text_lens), max_mel, max_text)
                    dur_padded.zero_()
                    for i in range(len(data[key])):
                        dur = data[key][i]
                        dur_padded[i, : dur.size(0), : dur.size(1)] = dur
                    data[key] = dur_padded
                else:
                    data[key] = pad_sequence(
                        data[key], batch_first=True, padding_value=0
                    )
            if isinstance(data[key][0], int):
                data[key] = torch.IntTensor(data[key])

        data["src_lens"] = text_lens
        data["max_src_len"] = max_text
        data["mel_lens"] = mel_lens
        data["max_mel_len"] = max_mel
        return data

    def load_dataset(self):
        self.train_dataset = self.config.training.filelist_loader(
            self.config.training.training_filelist
        )
        self.val_dataset = self.config.training.filelist_loader(
            self.config.training.validation_filelist
        )

    def prepare_data(self):
        if self.inference:
            self.predict_dataset = FastSpeechDataset(
                self.predict_dataset,
                self.config,
                self.lang2id,
                self.speaker2id,
                inference=self.inference,
                teacher_forcing=self.teacher_forcing,
            )
            torch.save(self.predict_dataset, self.predict_path)
        elif not self.prepared:
            self.train_dataset = (
                filter_dataset_based_on_target_text_representation_level(
                    self.config.model.target_text_representation_level,
                    self.train_dataset,
                    "training",
                    self.batch_size,
                )
            )
            self.val_dataset = filter_dataset_based_on_target_text_representation_level(
                self.config.model.target_text_representation_level,
                self.val_dataset,
                "validation",
                self.batch_size,
            )
            train_samples = len(self.train_dataset)
            val_samples = len(self.val_dataset)
            check_dataset_size(self.batch_size, train_samples, "training")
            check_dataset_size(self.batch_size, val_samples, "validation")
            self.train_dataset = FastSpeechDataset(
                self.train_dataset,
                self.config,
                self.lang2id,
                self.speaker2id,
                inference=self.inference,
            )
            self.val_dataset = FastSpeechDataset(
                self.val_dataset,
                self.config,
                self.lang2id,
                self.speaker2id,
                inference=self.inference,
            )
            # save it to disk
            torch.save(self.train_dataset, self.train_path)
            torch.save(self.val_dataset, self.val_path)
            self.prepared = True


class FastSpeech2SynthesisDataModule(FastSpeech2DataModule):
    def __init__(
        self,
        config: FastSpeech2Config,
        data: list[dict],
        lang2id: LookupTable,
        speaker2id: LookupTable,
        teacher_forcing: bool = False,
    ):
        super().__init__(config=config, inference=True, teacher_forcing=teacher_forcing)
        self.inference = True
        self.data = data
        self.collate_fn = partial(
            self.collate_method, learn_alignment=config.model.learn_alignment
        )
        self.use_weighted_sampler = config.training.use_weighted_sampler
        self.batch_size = config.training.batch_size
        self.load_dataset(data)
        self.predict_dataset_length = len(self.predict_dataset)
        self.lang2id = lang2id
        self.speaker2id = speaker2id

    def load_dataset(self, data):
        self.predict_dataset = data



import os
from functools import partial

import gradio as gr
import torch

from everyvoice.config.type_definitions import TargetTrainingTextRepresentationLevel
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.cli.synthesize import (
    synthesize_helper,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.model import (
    FastSpeech2,
)
from everyvoice.model.feature_prediction.FastSpeech2_lightning.fs2.prediction_writing_callback import (
    PredictionWritingWavCallback,
)
from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.utils import (
    load_hifigan_from_checkpoint,
)
from everyvoice.utils.heavy import get_device_from_accelerator

os.environ["no_proxy"] = "localhost,127.0.0.1,::1"


def synthesize_audio(
    text,
    duration_control,
    language,
    speaker,
    text_to_spec_model,
    vocoder_model,
    vocoder_config,
    accelerator,
    device,
    output_dir=None,
):
    if text == "":
        raise gr.Error(
            "Text for synthesis was not provided. Please type the text you want to be synthesized into the textfield."
        )
    if language is None:
        raise gr.Error("Language is not selected. Please select a language.")
    if speaker is None:
        raise gr.Error("Speaker is not selected. Please select a speaker.")
    config, device, predictions = synthesize_helper(
        model=text_to_spec_model,
        vocoder_model=vocoder_model,
        vocoder_config=vocoder_config,
        texts=[text],
        language=language,
        accelerator=accelerator,
        devices="1",
        device=device,
        global_step=1,
        vocoder_global_step=1,  # dummy value since the vocoder step is not used
        output_type=[],
        text_representation=TargetTrainingTextRepresentationLevel.characters,
        output_dir=output_dir,
        speaker=speaker,
        duration_control=duration_control,
        filelist=None,
        teacher_forcing_directory=None,
        batch_size=1,
        num_workers=1,
    )
    output_key = (
        "postnet_output" if text_to_spec_model.config.model.use_postnet else "output"
    )
    wav_writer = PredictionWritingWavCallback(
        output_dir=output_dir,
        config=config,
        output_key=output_key,
        device=device,
        global_step=1,
        vocoder_global_step=1,  # dummy value since the vocoder step is not used
        vocoder_model=vocoder_model,
        vocoder_config=vocoder_config,
    )
    # move to device because lightning accumulates predictions on cpu
    predictions[0][output_key] = predictions[0][output_key].to(device)
    wav, sr = wav_writer.synthesize_audio(predictions[0])
    return sr, wav[0]


def create_demo_app(
    text_to_spec_model_path,
    spec_to_wav_model_path,
    output_dir,
    accelerator,
) -> gr.Blocks:
    device = get_device_from_accelerator(accelerator)
    vocoder_ckpt = torch.load(spec_to_wav_model_path, map_location=device)
    vocoder_model, vocoder_config = load_hifigan_from_checkpoint(vocoder_ckpt, device)
    model: FastSpeech2 = FastSpeech2.load_from_checkpoint(text_to_spec_model_path).to(
        device
    )
    model.eval()
    synthesize_audio_preset = partial(
        synthesize_audio,
        text_to_spec_model=model,
        vocoder_model=vocoder_model,
        vocoder_config=vocoder_config,
        output_dir=output_dir,
        accelerator=accelerator,
        device=device,
    )
    lang_list = list(model.lang2id.keys())
    speak_list = list(model.speaker2id.keys())
    with gr.Blocks() as demo:
        gr.Markdown(
            """
            <h1 align="center">EveryVoice Demo</h1>
            """
        )
        with gr.Row():
            with gr.Column():
                inp_text = gr.Text(
                    placeholder="This text will be turned into speech.",
                    label="Input Text",
                )
                inp_slider = gr.Slider(
                    0.75, 1.75, 1.0, step=0.25, label="Duration Multiplier"
                )
                with gr.Row():
                    inp_lang = gr.Dropdown(lang_list, label="Language")
                    inp_speak = gr.Dropdown(speak_list, label="Speaker")
                btn = gr.Button("Synthesize")
            with gr.Column():
                out_audio = gr.Audio(format="mp3")
        btn.click(
            synthesize_audio_preset,
            inputs=[inp_text, inp_slider, inp_lang, inp_speak],
            outputs=[out_audio],
        )
    return demo



import sys
import textwrap
from pathlib import Path
from typing import Any, Optional

import typer
from everyvoice.base_cli.interfaces import complete_path
from everyvoice.config.type_definitions import (
    DatasetTextRepresentation,
    TargetTrainingTextRepresentationLevel,
)
from loguru import logger

from ..type_definitions import SynthesizeOutputFormats
from ..utils import truncate_basename


def validate_data_keys_with_model_keys(
    data_keys: set[str], model_keys: set[str], key: str, multi: bool
) -> None:
    """
    Make sure the user-supplied language or speaker specification is compatible with the model.
    In current version of model, we can validate either key="language" or key="speaker"
    Args:
        data_keys: values for key found in the data
        model_keys: values for key found in the model
        key: "language" or "speaker"
        multi: whether the model was trained in multilingual/speaker mode
    """
    if multi:
        if None in data_keys:
            print(
                f"You have not specified a {key} for all your sentences."
                f" Available values are {model_keys}",
                file=sys.stderr,
            )
            sys.exit(1)

        extras = data_keys.difference(model_keys)
        if extras:
            is_or_are_not = (
                f"are not {key}s that are" if len(data_keys) > 1 else f"is not a {key}"
            )
            print(
                f"You provided {data_keys} which {is_or_are_not} supported by the model {model_keys or {}}.",
                file=sys.stderr,
            )
            sys.exit(1)
    else:
        # NOTE: Even in non multiX, the model has a default value.
        # Looking at a filelist.psv
        # basename|characters|language|speaker
        # LJ002-0234|In the yard behind the prison|und|default
        # TODO: Instead, should we check that `data_keys == model_keys`?
        extras = data_keys.difference(model_keys | {None})
        if extras:
            print(
                f"The current model doesn't support multiple {key}s but your data has {key}s {extras}.\n"
                f"Please retrain your model with multi{'lingual' if key == 'language' else key} set to True.",
                file=sys.stderr,
            )
            sys.exit(1)


def prepare_data(
    texts: list[str],
    language: str | None,
    speaker: str | None,
    filelist: Path,
    # model is of type ..model.FastSpeech2, but we make it Any to keep the CLI
    # fast and enable mocking in unit testing.
    model: Any,
    text_representation: DatasetTextRepresentation,
    duration_control: float,
) -> list[dict[str, Any]]:
    """"""
    from everyvoice.utils import slugify

    data: list[dict[str, Any]]
    # NOTE: The wizard adds a default speaker=`default` to the data.
    # It also asks for a language that the user choses from a list which then becomes the default lanaguage, like `und`.
    # Knowing this, model.*2id should always have a default value thus DEFAULT_* should never be `None`.
    DEFAULT_LANGUAGE = next(iter(model.lang2id.keys()), None)
    DEFAULT_SPEAKER = next(iter(model.speaker2id.keys()), None)
    if texts:
        print(f"Processing text {texts}", file=sys.stderr)
        data = [
            {
                "basename": truncate_basename(slugify(text)),
                text_representation.value: text,
                "language": language or DEFAULT_LANGUAGE,
                "speaker": speaker or DEFAULT_SPEAKER,
            }
            for text in texts
        ]
    else:
        data = model.config.training.filelist_loader(filelist)
        try:
            data = [
                d
                | {
                    "basename": d.get(
                        "basename",
                        truncate_basename(slugify(d[text_representation.value])),
                    ),  # Only truncate the basename if the basename doesn't already exist in the filelist.
                    "language": language or d.get("language", DEFAULT_LANGUAGE),
                    "speaker": speaker or d.get("speaker", DEFAULT_SPEAKER),
                }
                for d in data
            ]
        except KeyError:
            # TODO: Errors should have better formatting:
            #       https://github.com/roedoejet/FastSpeech2_lightning/issues/26
            logger.info(
                textwrap.dedent(
                    """
                EveryVoice only accepts filelists in PSV format as in:

                    basename|characters|language|speaker
                    LJ0001|Hello|eng|LJ

                Or in a format where each new line is an utterance:

                    This is a sentence.
                    Here is another sentence.

                Your filelist did not contain the correct keys so we will assume it is in the plain text format.
                Text can either be defined as 'characters' or 'phones'.
                        """
                )
            )
            with open(filelist, encoding="utf8") as f:
                data = [
                    {
                        "basename": truncate_basename(slugify(line.strip())),
                        text_representation.value: line.strip(),
                        "language": language or DEFAULT_LANGUAGE,
                        "speaker": speaker or DEFAULT_SPEAKER,
                    }
                    for line in f
                ]

    validate_data_keys_with_model_keys(
        data_keys=set(d["language"] for d in data),
        model_keys=set(model.lang2id.keys()),
        key="language",
        multi=model.config.model.multilingual,
    )
    validate_data_keys_with_model_keys(
        data_keys=set(d["speaker"] for d in data),
        model_keys=set(model.speaker2id.keys()),
        key="speaker",
        multi=model.config.model.multispeaker,
    )

    # Add duration_control
    for item in data:
        item["duration_control"] = duration_control

    return data


def get_global_step(model_path: Path) -> int:
    """
    Extract the `global_step` from a model.
    Note: we have to do this because the `global_step` gets reset to 0 when we call load_from_checkpoint().
    """
    import torch

    m = torch.load(model_path, map_location=torch.device("cpu"))
    return m["global_step"]


def synthesize_helper(
    model,
    vocoder_model,
    vocoder_config,
    texts: list[str],
    language: Optional[str],
    speaker: Optional[str],
    duration_control: Optional[float],
    global_step: int,
    vocoder_global_step: int,
    output_type: list[SynthesizeOutputFormats],
    text_representation: TargetTrainingTextRepresentationLevel,
    accelerator: str,
    devices: str,
    device,
    batch_size: int,
    num_workers: int,
    filelist: Path,
    output_dir: Path,
    teacher_forcing_directory: Path,
):
    """This is a helper to perform synthesis once the model has been loaded.
    It allows us to use the same command for synthesis via the CLI and
    via the gradio demo.
    """
    from everyvoice.text.phonemizer import AVAILABLE_G2P_ENGINES

    from ..dataset import FastSpeech2SynthesisDataModule

    if (
        model.config.model.target_text_representation_level
        == TargetTrainingTextRepresentationLevel.characters
        and text_representation != DatasetTextRepresentation.characters
    ):
        raise ValueError(
            f"Your model was trained on {model.config.model.target_text_representation_level} but you provided {text_representation.value} which is incompatible."
        )
    if (
        model.config.model.target_text_representation_level
        != TargetTrainingTextRepresentationLevel.characters
        and text_representation == DatasetTextRepresentation.characters
        and language not in AVAILABLE_G2P_ENGINES
    ):
        raise ValueError(
            f"Your model was trained on {model.config.model.target_text_representation_level} but you provided {text_representation.value} and there is no available grapheme-to-phoneme engine available for {language}. Please see <TODO: Docs!> for more information on how to add one."
        )

    data = prepare_data(
        texts=texts,
        language=language,
        speaker=speaker,
        duration_control=duration_control if duration_control else 1.0,
        filelist=filelist,
        model=model,
        text_representation=text_representation,
    )

    from pytorch_lightning import Trainer

    from ..prediction_writing_callback import get_synthesis_output_callbacks

    trainer = Trainer(
        logger=False,  # We don't need to log things to tensorboard during inference
        accelerator=accelerator,
        devices=devices,
        max_epochs=model.config.training.max_epochs,
        callbacks=get_synthesis_output_callbacks(
            output_type=output_type,
            output_dir=output_dir,
            config=model.config,
            output_key=model.output_key,
            device=device,
            global_step=global_step,
            vocoder_model=vocoder_model,
            vocoder_config=vocoder_config,
            vocoder_global_step=vocoder_global_step,
        ),
    )
    if teacher_forcing_directory is not None:
        teacher_forcing = True
        model.config.preprocessing.save_dir = teacher_forcing_directory
    else:
        teacher_forcing = False
    # overwrite batch_size and num_workers
    model.config.training.batch_size = batch_size
    model.config.training.train_data_workers = num_workers
    return (
        model.config,
        device,
        trainer.predict(
            model,
            FastSpeech2SynthesisDataModule(
                model.config,
                data,
                model.lang2id,
                model.speaker2id,
                teacher_forcing=teacher_forcing,
            ),
            return_predictions=True,
        ),
    )


def synthesize(  # noqa: C901
    model_path: Path = typer.Argument(
        ...,
        file_okay=True,
        exists=True,
        dir_okay=False,
        help="The path to a trained text-to-spec or e2e EveryVoice model.",
        autocompletion=complete_path,
    ),
    output_dir: Path = typer.Option(
        "synthesis_output",
        "--output-dir",
        "-o",
        file_okay=False,
        dir_okay=True,
        help="The directory where your synthesized audio should be written",
        autocompletion=complete_path,
    ),
    texts: list[str] = typer.Option(
        [],
        "--text",
        "-t",
        help="Some text to synthesize.  This option can be repeated to synthesize multiple sentences."
        " It is recommended to use --filelist if you want to synthesize a lot of sentences or have different speaker/language per sentence.",
    ),
    language: Optional[str] = typer.Option(
        None,
        "--language",
        "-l",
        help="Specify which language to use in a multilingual system. [requires --text]",
    ),
    duration_control: Optional[float] = typer.Option(
        1.0,
        "--duration-control",
        "-D",
        help="Control the speaking rate of the synthesis. Set a value to multily the durations by, lower numbers produce quicker speaking rates, larger numbers produce slower speaking rates. Default is 1.0",
    ),
    speaker: Optional[str] = typer.Option(
        None,
        "--speaker",
        "-s",
        help="Specify which speaker to use in a multispeaker system. [requires --text]",
    ),
    accelerator: str = typer.Option("auto", "--accelerator", "-a"),
    devices: str = typer.Option(
        "auto", "--devices", "-d", help="The number of GPUs to use"
    ),
    filelist: Path = typer.Option(
        None,
        "--filelist",
        "-f",
        exists=True,
        file_okay=True,
        dir_okay=False,
        help="Synthesize all audio in a given filelist. Use --text if you want to just synthesize one sample.",
        autocompletion=complete_path,
    ),
    text_representation: DatasetTextRepresentation = typer.Option(
        DatasetTextRepresentation.characters,
        help="The representation of the text you are synthesizing. Can be either 'characters', 'phones', or 'arpabet'. The input type must be compatible with your model.",
    ),
    output_type: list[SynthesizeOutputFormats] = typer.Option(
        [SynthesizeOutputFormats.wav.value],
        "-O",
        "--output-type",
        help="""Which format(s) to synthesize to.
        Multiple formats can be provided by repeating `--output-type`.
        **wav** is the default and will synthesize to a playable audio file;
        **spec** will generate predicted Mel spectrograms. Tensors are time-oriented (T, K) where T is equal to the number of frames and K is equal to the number of Mel bands.
        **textgrid** will generate a Praat TextGrid with alignment labels. This can be helpful for evaluation.
        """,
    ),
    teacher_forcing_directory: Path = typer.Option(
        None,
        "--teacher-forcing-directory",
        "-T",
        help="ADVANCED. The path to preprocessed folder containing spec and duration folders to use for teacher-forcing the synthesized outputs.",
        dir_okay=True,
        file_okay=False,
        autocompletion=complete_path,
    ),
    vocoder_path: Path = typer.Option(
        None,
        "--vocoder-path",
        "-v",
        help="The path to a trained vocoder (aka spec-to-wav model).",
        dir_okay=False,
        file_okay=True,
        autocompletion=complete_path,
    ),
    batch_size: int = typer.Option(
        4,
        "--batch-size",
        "-b",
        help="Batch size.",
    ),
    num_workers: int = typer.Option(
        4,
        "--num-workers",
        "-n",
        help="Number of workers to process the data.",
    ),
):

    """Given some text and a trained model, generate some audio. i.e. perform typical speech synthesis"""
    # TODO: allow for changing of language/speaker and variance control
    import torch
    from everyvoice.model.vocoder.HiFiGAN_iSTFT_lightning.hfgl.utils import (
        load_hifigan_from_checkpoint,
    )
    from everyvoice.utils.heavy import get_device_from_accelerator

    from ..model import FastSpeech2

    if texts and filelist:
        print(
            "Got arguments for both text and a filelist - this will only process the text."
            " Please re-run without providing text if you want to run batch synthesis on the provided file.",
            file=sys.stderr,
        )
    if not texts and not filelist:
        print("You must define either --text or --filelist", file=sys.stderr)
        sys.exit(1)

    # output to .wav will require a valid spec-to-wav model
    if SynthesizeOutputFormats.wav in output_type and not vocoder_path:
        print(
            "Missing --vocoder-path option, which is required when the output type is 'wav'.",
            file=sys.stderr,
        )
        sys.exit(1)

    output_dir.mkdir(exist_ok=True, parents=True)
    # NOTE: We want to be able to put the vocoder on the proper accelerator for
    # it to be compatible with the vocoder's input device.
    # We could misuse the trainer's API and use the private variable
    # trainer._accelerator_connector._accelerator_flag but that value is
    # computed when instantiating a trainer and that is exactly when we need
    # the information to create the callbacks.
    device = get_device_from_accelerator(accelerator)

    # Load checkpoints
    print(f"Loading checkpoint from {model_path}", file=sys.stderr)
    model: FastSpeech2 = FastSpeech2.load_from_checkpoint(model_path).to(device)
    model.eval()

    # get global step
    # We can't just use model.global_step because it gets reset by lightning
    global_step = get_global_step(model_path)

    # load vocoder
    logger.info(f"Loading Vocoder from {vocoder_path}")
    if vocoder_path is None:
        logger.error(
            "No vocoder was provided, please specify "
            "--vocoder-path /path/to/vocoder on the command line."
        )
        sys.exit(1)
    else:
        vocoder_ckpt = torch.load(vocoder_path, map_location=device)
        vocoder_model, vocoder_config = load_hifigan_from_checkpoint(
            vocoder_ckpt, device
        )
        # We can't just use model.global_step because it gets reset by lightning
        vocoder_global_step = get_global_step(vocoder_path)
        return synthesize_helper(
            model=model,
            texts=texts,
            language=language,
            speaker=speaker,
            duration_control=duration_control,
            global_step=global_step,
            output_type=output_type,
            text_representation=text_representation,
            accelerator=accelerator,
            devices=devices,
            device=device,
            batch_size=batch_size,
            num_workers=num_workers,
            filelist=filelist,
            teacher_forcing_directory=teacher_forcing_directory,
            output_dir=output_dir,
            vocoder_model=vocoder_model,
            vocoder_config=vocoder_config,
            vocoder_global_step=vocoder_global_step,
        )



from enum import Enum
from pathlib import Path

import typer
from everyvoice.base_cli.interfaces import (
    complete_path,
    preprocess_base_command_interface,
    train_base_command_interface,
)
from loguru import logger
from merge_args import merge_args
from tqdm import tqdm

from .config import DFAlignerConfig

app = typer.Typer(
    pretty_exceptions_show_locals=False,
    context_settings={"help_option_names": ["-h", "--help"]},
    help="A fork of the DeepForcedAligner project implemented in PyTorch Lightning",
)


class PreprocessCategories(str, Enum):
    audio = "audio"
    spec = "spec"
    text = "text"


@app.command()
@merge_args(preprocess_base_command_interface)
def preprocess(
    steps: list[PreprocessCategories] = typer.Option(
        [cat.value for cat in PreprocessCategories],
        "-s",
        "--steps",
        help="Which steps of the preprocessor to use. If none are provided, all steps will be performed.",
    ),
    **kwargs,
):
    from everyvoice.base_cli.helpers import preprocess_base_command

    preprocess_base_command(
        model_config=DFAlignerConfig,
        steps=[step.name for step in steps],
        **kwargs,
    )


@app.command()
@merge_args(train_base_command_interface)
def train(**kwargs):
    from everyvoice.base_cli.helpers import train_base_command

    from .dataset import AlignerDataModule
    from .model import Aligner

    train_base_command(
        model_config=DFAlignerConfig,
        model=Aligner,
        data_module=AlignerDataModule,
        monitor="validation/loss",
        gradient_clip_val=1.0,
        **kwargs,
    )


@app.command()
def extract_alignments(
    config_file: Path = typer.Argument(
        ...,
        exists=True,
        dir_okay=False,
        file_okay=True,
        help="The path to your model configuration file.",
        autocompletion=complete_path,
    ),
    accelerator: str = typer.Option("auto", "--accelerator", "-a"),
    devices: str = typer.Option("auto", "--devices", "-d"),
    model_path: Path = typer.Option(
        None,
        "--model",
        "-m",
        exists=True,
        file_okay=True,
        dir_okay=False,
        autocompletion=complete_path,
    ),
    config_args: list[str] = typer.Option(None, "--config", "-c"),
    num_processes: int = typer.Option(None),
    predict: bool = typer.Option(True),
    create_n_textgrids: int = typer.Option(5, "--tg", "--n_textgrids"),
):
    from everyvoice.utils import update_config_from_cli_args

    from .utils import create_textgrid, extract_durations_for_item

    config = DFAlignerConfig.load_config_from_path(config_file)

    config = update_config_from_cli_args(config_args, config)

    # Imports
    logger.info("Loading modules for alignment...")
    pbar = tqdm(range(6))
    pbar.set_description("Loading pytorch and friends")
    import numpy as np

    pbar.update()
    import torch

    pbar.update()
    from pytorch_lightning.loggers import TensorBoardLogger

    pbar.update()
    from pytorch_lightning import Trainer

    pbar.update()
    pbar.set_description("Loading EveryVoice modules")
    from .dataset import AlignerDataModule

    pbar.update()
    from everyvoice.text.text_processor import TextProcessor

    from .model import Aligner

    pbar.update()

    # TODO: make this faster
    if num_processes is None:
        num_processes = 4

    data = AlignerDataModule(config)
    save_dir = Path(config.preprocessing.save_dir)
    (save_dir / "duration").mkdir(exist_ok=True, parents=True)
    if predict:
        tensorboard_logger = TensorBoardLogger(
            **{
                **(
                    config.training.logger.model_dump(
                        exclude={"sub_dir_callable": True}
                    )
                ),
                **{"sub_dir": config.training.logger.sub_dir},
            }
        )
        trainer = Trainer(
            accelerator=accelerator, devices=devices, logger=tensorboard_logger
        )
        if model_path:
            model: Aligner = Aligner.load_from_checkpoint(
                model_path.absolute().as_posix()
            )
            # TODO: check into the best way to update config from re-loaded model
            # model.update_config(config)
            model.config = config
            trainer.predict(model, dataloaders=data)
        else:
            trainer.predict(dataloaders=data)
    SEP = "--"
    tg_processed = 0
    text_processor = TextProcessor(config.text)
    for item in tqdm(
        data.predict_dataloader().dataset,
        total=len(data.predict_dataloader().dataset),
    ):
        basename = item["basename"]
        speaker = item["speaker"]
        language = item["language"]
        tokens = item["tokens"].cpu()
        pred = np.load(
            save_dir
            / "duration"
            / SEP.join([basename, speaker, language, "duration.npy"])
        )
        item, durations = extract_durations_for_item(
            item,
            tokens,
            pred,
            # ignore mypy type checking because https://github.com/pydantic/pydantic/issues/3809
            method=config.training.extraction_method,  # type: ignore
        )
        assert len(durations) == len(
            tokens
        ), f"Length of tokens and durations must be the same, but was not for {basename}. Try re-running with dijkstra extraction method. This might be because your model was not trained properly or because you are showing it unseen data."
        if tg_processed < create_n_textgrids:
            (save_dir / "text_grid").mkdir(parents=True, exist_ok=True)
            create_textgrid(
                save_dir
                / "text_grid"
                / SEP.join([basename, speaker, language, "duration.TextGrid"]),
                text_processor.token_sequence_to_text_sequence(tokens.tolist()),
                durations,
                hop_size=config.preprocessing.audio.fft_hop_size,
                sample_rate=config.preprocessing.audio.alignment_sampling_rate,
            )
            tg_processed += 1
        torch.save(
            torch.IntTensor(durations),
            save_dir
            / "duration"
            / SEP.join([basename, speaker, language, "duration.pt"]),
        )


if __name__ == "__main__":
    app()
